#!/usr/bin/env python
# install_wizard.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (install_wizard.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: install_wizard.

.. raw:: html

    <a href="https://bitdust.io/automats/install_wizard/install_wizard.png" target="_blank">
    <img src="https://bitdust.io/automats/install_wizard/install_wizard.png" style="max-width:100%;">
    </a>

A state machine to show installation wizard for BitDust software.
User need to answer some questions step by step to configure the program first time.

This is several pages:
    * select a role: "free backups", "secure storage", "donator", "just to try", "beta tester"
    * set up needed and donated space and local folders locations
    * provide some personal information about yourself if you wish
    * set software update settings

EVENTS:
    * :red:`back`
    * :red:`next`
    * :red:`skip`
"""

from __future__ import absolute_import
import sys
try:
    pass
except:
    sys.exit('Error initializing twisted.internet.reactor in install_wizard.py')

from bitdust.main import config
from bitdust.automats import automat

from bitdust.main import installer

#------------------------------------------------------------------------------

_InstallWizard = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _InstallWizard
    if _InstallWizard is None:
        _InstallWizard = InstallWizard('install_wizard', 'READY', 2)
    if event is not None:
        _InstallWizard.automat(event, *args, **kwargs)
    return _InstallWizard


class InstallWizard(automat.Automat):
    """
    BitDust install_wizard() Automat.

    Runs install wizard process.
    """

    fast = True

    role_args = None
    role = None

    def init(self):
        self.log_events = True
        # TODO: we do not need 'READY' state now
        # because only have one page "STORAGE"
        self.state = 'STORAGE'

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        if newstate == 'CONTACTS' and oldstate == 'STORAGE':
            self.event('next', {})
            # TODO:
            # here just skip Contacts page!
            # we do not need that now, but can back to that soon when add chat
        installer.A('install_wizard.state', newstate)

    def A(self, event, *args, **kwargs):
        #---READY---
        if self.state == 'READY':
            if event == 'next':
                self.state = 'STORAGE'
            elif event == 'skip':
                self.state = 'LAST_PAGE'
        #---STORAGE---
        elif self.state == 'STORAGE':
            if event == 'next':
                self.state = 'CONTACTS'
                self.doSaveStorage(*args, **kwargs)
            elif event == 'back':
                self.state = 'READY'
        #---CONTACTS---
        elif self.state == 'CONTACTS':
            if event == 'back':
                self.state = 'STORAGE'
            elif event == 'next':
                self.state = 'LAST_PAGE'
                self.doSaveContacts(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---LAST_PAGE---
        elif self.state == 'LAST_PAGE':
            if event == 'next':
                self.state = 'DONE'
            elif event == 'back':
                self.state = 'CONTACTS'
        return None

    def doSaveStorage(self, *args, **kwargs):
        needed = args[0].get('needed', '')
        donated = args[0].get('donated', '')
        customersdir = args[0].get('customersdir', '')
        localbackupsdir = args[0].get('localbackupsdir', '')
        restoredir = args[0].get('restoredir', '')
        if needed:
            config.conf().setData('services/customer/needed-space', needed + ' MB')
        if donated:
            config.conf().setData('services/supplier/donated-space', donated + ' MB')
        if customersdir:
            config.conf().setString('paths/customers', customersdir)
        if localbackupsdir:
            config.conf().setString('paths/backups', localbackupsdir)
        if restoredir:
            config.conf().setString('paths/restore', restoredir)

    def doSaveContacts(self, *args, **kwargs):
        pass



import sqlite3, time, options, sys

import tornado.ioloop
import tornado.web

config = options.Get()
config.read()
hyper_path = config.hyper_path
#hyper_path = "backup.db"
version = config.version

if 'testnet' in version:
    port = 2829
    hyper_path = 'static/ledger_test.db'


def execute(cursor, query):
    """Secure execute for slow nodes"""
    while True:
        try:
            cursor.execute(query)
            break
        except Exception as e:
            print('Database query: {} {}'.format(cursor, query))
            print('Database retry reason: {}'.format(e))
    return cursor


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        # redraw chart

        #if full_ledger == 1:
        #    conn = sqlite3.connect(ledger_path)
        #else:

        conn = sqlite3.connect(hyper_path, timeout=60.0)

        c = conn.cursor()
        execute(c, 'SELECT * FROM transactions ORDER BY block_height DESC, timestamp DESC LIMIT 100;')
        all = c.fetchall()[::-1]

        axis0 = []
        axis1 = []
        axis4 = []
        axis8 = []
        axis9 = []
        axis10 = []

        i = 1
        for x in all:
            axis0.append(x[0])  # append block height
            axis1.append(x[1])  # append timestamp
            axis4.append(x[4])  # append amount
            axis8.append(x[8])  # append fee
            axis9.append(x[9])  # append reward
            axis10.append(len(str(x)))  # tx size

        plotter = []

        # define canvas
        plotter.append('<canvas id="canvas" height="150" width="600"></canvas>\n')
        # define canvas

        plotter.append('<script>\n')

        # onload
        plotter.append("var ctx = document.getElementById('canvas').getContext('2d');")
        # onload

        # segment
        plotter.append('var canvas = new Chart(ctx, {\n')
        plotter.append("type: 'line',\n")
        plotter.append('data: {\n')
        plotter.append('labels: ' + str(list(map(str, axis0))) + ',\n')
        plotter.append('datasets: [{\n')
        plotter.append("label: 'Timestamp progression',\n")
        plotter.append('data: ' + str(list(map(str, axis1))) + ',\n')
        plotter.append("backgroundColor: 'rgba(153, 255, 51, 0.4)'\n")
        plotter.append('}, {\n')
        plotter.append("label: 'Spending',\n")
        plotter.append('hidden: true,\n')
        plotter.append('data: ' + str(list(map(str, axis4))) + ',\n')
        plotter.append("backgroundColor: 'rgba(255, 153, 0, 0.4)'\n")
        plotter.append('}, {\n')
        plotter.append("label: 'Fee',\n")
        plotter.append('hidden: true,\n')
        plotter.append('data: ' + str(list(map(str, axis8))) + ',\n')
        plotter.append("backgroundColor: 'rgba(63, 65, 191, 0.4)'\n")
        plotter.append('}, {\n')

        plotter.append("label: 'Transaction size',\n")
        plotter.append('hidden: true,\n')
        plotter.append('data: ' + str(list(map(str, axis10))) + ',\n')
        plotter.append("backgroundColor: 'rgba(300, 50, 0, 0.4)'\n")
        plotter.append('}, {\n')

        plotter.append("label: 'Reward',\n")
        plotter.append('hidden: true,\n')
        plotter.append('data: ' + str(list(map(str, axis9))) + ',\n')
        plotter.append("backgroundColor: 'rgba(189, 63, 191, 0.4)'\n")
        plotter.append('}]\n')
        plotter.append('}\n')
        plotter.append('});\n')
        # segment

        plotter.append('</script>\n')
        # plotter.append('</body>\n')
        # plotter.append('</html>')
        # redraw chart

        execute(c, 'SELECT * FROM transactions ORDER BY block_height DESC, timestamp DESC LIMIT 500;')
        all = c.fetchall()

        tx_count = 0
        for x in all:
            if x[9] == 0:  #if reward is 0
                tx_count = tx_count + 1

        transferred_total = 0
        for x in all:
            if x[9] == 0:  # if reward is 0
                transferred_total = transferred_total + x[4]

        execute(c, 'SELECT difficulty FROM misc ORDER BY block_height DESC LIMIT 500;')
        diffs = c.fetchall()

        view = []
        i = 0
        b = -1

        for x in all:

            try:  #first run
                x_old
            except:
                x_old = 'init'

            if x[0] != x_old:
                color_cell = '#E8E8E8'
                view.append('<tr><td></td></tr>')  #block separator
            else:
                color_cell = 'white'

            view.append('<tr bgcolor ={}>'.format(color_cell))

            if x[0] != x_old:
                b = b + 1

            if x_old != x[0]:
                view.append('<td>{}</td>'.format(x[0]))  #block height
            else:
                view.append('<td></td>')

            view.append('<td>{}'.format(time.strftime('%Y/%m/%d,%H:%M:%S', time.gmtime(float(x[1])))))
            view.append('<td>{}</td>'.format(x[2]))
            view.append('<td>{}</td>'.format(x[3]))
            view.append('<td>{}</td>'.format('%.8f' % float(x[4])))
            view.append('<td>{}</td>'.format(x[5][:56]))

            if x_old != x[0]:
                view.append('<td>{}</td>'.format(x[7]))  #block hash
            else:
                view.append('<td></td>')

            view.append('<td>{}</td>'.format(x[8]))
            view.append('<td>{}</td>'.format('%.6f' % float(x[9])))

            if x_old != x[0]:
                view.append('<td>{}</td>'.format('%.10f' % float(diffs[b][0])))
            else:
                view.append('<td></td>')

            view.append('<tr>')

            x_old = x[0]
            i = i + 1

        c.close()

        html = []
        html.append('<!doctype html>\n')
        html.append('<html>\n')

        html.append('<head>\n')
        # plotter.append('<meta http-equiv="refresh" content="60" >')

        html.append('<title>Transaction Explorer</title>\n')
        # html.append('<link rel="stylesheet" type="text/css" href="static/style.css">')
        html.append('<link rel = "icon" href = "static/explorer.ico" type = "image/x-icon" / >\n')
        html.append('<link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" >')
        html.append('<script src="static/Chart.js"></script>\n')
        html.append('<script src="https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js"></script>')
        html.append('<script src="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/js/bootstrap.min.js"></script >')

        html.append('</head>\n')
        html.append('<body bgcolor = "white">\n')
        html.append('<body>')
        html.append("<body background='static/explorer_bg.png'>")

        html.append("<div class ='container-fluid'>")

        html.append("<div class='row'>")
        html.append('<center><h1>Bismuth Transaction Explorer</h1><p></center>')

        html.append("<div style='padding: 20px;'>")
        html.append("<center><a href='http://bismuth.cz/ledger.tar.gz' class='btn btn-info' role='button'>Download Blockchain</a></center>")
        html.append('</div>')

        html.append('</div>')

        html.append("<div class ='row'>")
        html.append(''.join(plotter))
        html.append('</div>')

        data_total = str(sys.getsizeof(str(all))/1024)
        html.append("<div class ='container-fluid'>")
        html.append("<table class='table table-responsive'>")

        html.append('<tr><th>Statistics for the last 500 blocks</th>')
        html.append('<tr><td>Size in KB: </td><td>{}</td>'.format(data_total))
        html.append('<tr><td>Transactions: </td><td>{}</td>'.format(tx_count))
        html.append('<tr><td>Transactions per block: </td><td>{}</td>'.format(tx_count/500))
        html.append('<tr><td>Total BIS transferred: </td><td>{}</td>'.format(transferred_total))

        html.append('</table>')
        html.append('</div>')

        html.append("<div class ='row'>")
        html.append("<table class='table table-responsive'>")
        html.append("<tr bgcolor='white'>")

        html.append('<td>Block</td>')
        html.append('<td>Timestamp</td>')
        html.append('<td>From</td>')
        html.append('<td>To</td>')
        html.append('<td>Amount</td>')
        html.append('<td>TXID</td>')
        html.append('<td>Block Hash</td>')
        html.append('<td>Fee</td>')
        html.append('<td>Reward</td>')
        html.append('<td>Difficulty</td>')
        html.append('</tr>')
        html.append(''.join(view))
        html.append('</table>')
        html.append('</div>')

        html.append('</div>')

        html.append('</body>')
        html.append('</html>')

        self.write(''.join(html))
        #self.render("ex.html", data=data)


def make_app():

    return tornado.web.Application([
        (r'/', MainHandler),
        (
            r'/static/(.*)',
            tornado.web.StaticFileHandler,
            {
                'path': 'static',
            },
        ),
    ])


if __name__ == '__main__':
    app = make_app()
    app.listen(5492)
    tornado.ioloop.IOLoop.current().start()



#!/usr/bin/python
# service_customer_support.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_customer_support.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_customer_support
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return CustomerSupportService()


class CustomerSupportService(LocalService):

    service_name = 'service_customer_support'
    config_path = 'services/customer-support/enabled'

    def dependent_on(self):
        return [
            'service_customer_patrol',
        ]

    def start(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.userid import id_url
        from bitdust.main import events
        from bitdust.supplier import customer_assistant
        from bitdust.contacts import contactsdb
        from bitdust.transport import callback
        for customer_idurl in contactsdb.customers():
            if id_url.is_cached(customer_idurl):
                if customer_idurl and not customer_assistant.by_idurl(customer_idurl):
                    ca = customer_assistant.create(customer_idurl)
                    reactor.callLater(0, ca.automat, 'init')  # @UndefinedVariable
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        callback.add_outbox_callback(self._on_outbox_packet_sent)
        callback.append_inbox_callback(self._on_inbox_packet_received)
        return True

    def stop(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.main import events
        from bitdust.supplier import customer_assistant
        from bitdust.transport import callback
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        callback.remove_outbox_callback(self._on_outbox_packet_sent)
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        for ca in customer_assistant.assistants().values():
            reactor.callLater(0, ca.automat, 'shutdown')  # @UndefinedVariable
        return True

    def health_check(self):
        return True

    def _on_outbox_packet_sent(self, pkt_out):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.p2p import commands
        from bitdust.contacts import contactsdb
        from bitdust.supplier import customer_assistant
        if pkt_out.outpacket.Command == commands.Identity():
            if contactsdb.is_customer(pkt_out.outpacket.RemoteID):
                ca = customer_assistant.by_idurl(pkt_out.outpacket.RemoteID)
                if ca:
                    reactor.callLater(0, ca.automat, 'propagate', pkt_out)  # @UndefinedVariable
        return False

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.p2p import commands
        from bitdust.contacts import contactsdb
        from bitdust.supplier import customer_assistant
        if newpacket.Command in [commands.Ack(), commands.Fail()]:
            if contactsdb.is_customer(newpacket.OwnerID):
                ca = customer_assistant.by_idurl(newpacket.OwnerID)
                if ca:
                    reactor.callLater(0, ca.automat, newpacket.Command.lower(), newpacket)  # @UndefinedVariable
                    return True
        return False

    def _on_identity_url_changed(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.userid import id_url
        from bitdust.supplier import customer_assistant
        for customer_idurl, ca in customer_assistant.assistants().items():
            if customer_idurl == id_url.field(evt.data['old_idurl']):
                customer_idurl.refresh(replace_original=True)
                ca.customer_idurl.refresh(replace_original=True)
                lg.info('found %r to be refreshed after rotated identity: %r' % (ca, customer_idurl))
                reactor.callLater(0, ca.automat, 'connect')  # @UndefinedVariable



import base64, hashlib, json
from Cryptodome.PublicKey import RSA


def generate():
    # generate key pair and an address
    key = RSA.generate(4096)

    private_key_readable = key.exportKey().decode('utf-8')
    public_key_readable = key.publickey().exportKey().decode('utf-8')
    address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()  # hashed public key
    return private_key_readable, public_key_readable, address


def read():
    # import keys
    with open('wallet.der', 'r') as wallet_file:
        wallet_dict = json.load(wallet_file)
    private_key_readable = wallet_dict['Private Key']
    public_key_readable = wallet_dict['Public Key']
    key = private_key_readable

    if (len(public_key_readable)) != 271 and (len(public_key_readable)) != 799:
        raise ValueError('Invalid public key length: {}'.format(len(public_key_readable)))

    public_key_b64encoded = base64.b64encode(public_key_readable.encode('utf-8')).decode('utf-8')
    address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()
    # import keys

    return key, private_key_readable, public_key_readable, public_key_b64encoded, address



#!/usr/bin/env python
# memdebug.py
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (memdebug.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: memdebug.

This is to test memory usage. It was useful when I looked for memory
leaks. Can start a local HTTP server to keep track of all python objects
in the memory.
"""

from __future__ import absolute_import
import cherrypy  # @UnresolvedImport
import dowser  # @UnresolvedImport


def start(port):
    cherrypy.config.update({
        'environment': 'embedded',
        'server.socket_port': port,
        'server.socket_host': '0.0.0.0',
    })
    cherrypy.tree.mount(dowser.Root())

    # cherrypy.server.quickstart()
    cherrypy.engine.start()


def stop():
    cherrypy.engine.exit()


#from twisted.web.wsgi import WSGIResource
#from twisted.internet import reactor  # @UnresolvedImport
#
#from dozer import Dozer
#
# def application(environ, start_response):
#    start_response('200 OK', [('Content-type', 'text/plain')])
#    return ['Hello, world!']
#
# def start():
#    resource = WSGIResource(reactor, reactor.getThreadPool(), application)
#    wsgi_app = Dozer(application)



#!/usr/bin/env python
# supplier_connector.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (supplier_connector.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: supplier.

.. role:: red

BitDust supplier_connector() Automat

.. raw:: html

    <a href="supplier.png" target="_blank">
    <img src="supplier.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`ack`
    * :red:`connect`
    * :red:`disconnect`
    * :red:`fail`
    * :red:`queue-ack`
    * :red:`queue-fail`
    * :red:`queue-skip`
    * :red:`shutdown`
    * :red:`timer-10sec`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import math

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.main import settings
from bitdust.main import events

from bitdust.lib import strng
from bitdust.lib import nameurl
from bitdust.lib import diskspace
from bitdust.lib import jsn

from bitdust.contacts import contactsdb

from bitdust.services import driver

from bitdust.p2p import commands
from bitdust.p2p import p2p_service
from bitdust.p2p import online_status

from bitdust.raid import eccmap

from bitdust.storage import accounting

from bitdust.userid import id_url
from bitdust.userid import global_id
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_SuppliersConnectors = {}

#------------------------------------------------------------------------------


def connectors(customer_idurl=None, as_dict=False):
    global _SuppliersConnectors
    if as_dict:
        return _SuppliersConnectors
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _SuppliersConnectors:
        _SuppliersConnectors[customer_idurl] = {}
    return _SuppliersConnectors[customer_idurl]


def create(supplier_idurl, customer_idurl=None, needed_bytes=None, key_id=None, queue_subscribe=True):
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    assert supplier_idurl not in connectors(customer_idurl)
    connectors(customer_idurl)[supplier_idurl] = SupplierConnector(
        supplier_idurl=supplier_idurl,
        customer_idurl=customer_idurl,
        needed_bytes=needed_bytes,
        key_id=key_id,
        queue_subscribe=queue_subscribe,
    )
    return connectors(customer_idurl)[supplier_idurl]


def is_supplier(supplier_idurl, customer_idurl=None):
    global _SuppliersConnectors
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    if not id_url.is_cached(customer_idurl):
        return False
    if not id_url.is_cached(supplier_idurl):
        return False
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    if customer_idurl not in _SuppliersConnectors:
        return False
    if supplier_idurl not in _SuppliersConnectors[customer_idurl]:
        return False
    return True


def by_idurl(supplier_idurl, customer_idurl=None):
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    return connectors(customer_idurl).get(supplier_idurl, None)


def total_connectors():
    count = 0
    for suppliers_list in _SuppliersConnectors.values():
        count += len(suppliers_list)
    return count


#------------------------------------------------------------------------------


class SupplierConnector(automat.Automat):

    """
    This class implements all the functionality of the ``supplier_connector()``
    state machine.
    """

    timers = {
        'timer-30sec': (30.0, ['REQUEST']),
        'timer-10sec': (10.0, ['REFUSE', 'QUEUE?']),
    }

    def __init__(self, supplier_idurl, customer_idurl, needed_bytes, key_id=None, queue_subscribe=True):
        self.supplier_idurl = supplier_idurl
        self.supplier_id = self.supplier_idurl.to_id()
        self.customer_idurl = customer_idurl
        self.customer_id = self.customer_idurl.to_id()
        self.needed_bytes = needed_bytes
        self.key_id = key_id
        self.queue_subscribe = queue_subscribe
        self.do_calculate_needed_bytes()
        name = 'supplier_%s_%s_%s' % (
            nameurl.GetName(self.supplier_idurl),
            nameurl.GetName(self.customer_idurl),
            diskspace.MakeStringFromBytes(self.needed_bytes).replace(' ', ''),
        )
        self.request_packet_id = None
        self.request_queue_packet_id = None
        self.latest_supplier_ack = None
        self.callbacks = {}
        self.storage_contract = None
        try:
            st = bpio.ReadTextFile(settings.SupplierServiceFilename(
                supplier_idurl=self.supplier_idurl,
                customer_idurl=self.customer_idurl,
            )).strip()
        except:
            lg.exc()
        st = st or 'DISCONNECTED'
        automat.Automat.__init__(
            self,
            name,
            state=st,
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
        for cb_list in self.callbacks.values():
            for cb in cb_list:
                cb(self.supplier_idurl, self.state, self.state)

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self._last_known_family_position = None
        self._last_known_ecc_map = None
        self._last_known_family_snapshot = None
        self._supplier_connected_event_sent = False
        online_status.add_online_status_listener_callback(
            idurl=self.supplier_idurl,
            callback_method=self._on_online_status_state_changed,
        )

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when automat's state was changed.
        """
        if newstate in ['CONNECTED', 'DISCONNECTED', 'NO_SERVICE']:
            supplierPath = settings.SupplierPath(self.supplier_idurl, customer_idurl=self.customer_idurl)
            if not os.path.isdir(supplierPath):
                try:
                    os.makedirs(supplierPath)
                except:
                    lg.exc()
                    return
            bpio.WriteTextFile(
                settings.SupplierServiceFilename(supplier_idurl=self.supplier_idurl, customer_idurl=self.customer_idurl),
                newstate,
            )
        if newstate == 'CONNECTED':
            if id_url.is_the_same(self.customer_idurl, my_id.getIDURL()):
                #TODO: receive and process contract details: "pay_before" field
                pass
            if not self._supplier_connected_event_sent:
                self._supplier_connected_event_sent = True
                events.send('supplier-connected', data=dict(
                    supplier_idurl=self.supplier_idurl,
                    customer_idurl=self.customer_idurl,
                    needed_bytes=self.needed_bytes,
                    key_id=self.key_id,
                ))
        if newstate in ['DISCONNECTED', 'NO_SERVICE']:
            self._supplier_connected_event_sent = False

    def set_callback(self, name, cb):
        if name not in self.callbacks:
            self.callbacks[name] = []
        if cb in self.callbacks[name]:
            lg.warn('callback %r is already registered in %r with name %s' % (cb, self, name))
        self.callbacks[name].append(cb)

    def remove_callback(self, name, cb=None):
        if name in self.callbacks:
            if cb:
                while cb in self.callbacks[name]:
                    self.callbacks[name].remove(cb)
            else:
                self.callbacks.pop(name)
        else:
            lg.warn('callback with name %s not registered in %r' % (name, self))

    def do_calculate_needed_bytes(self):
        if self.needed_bytes is None:
            total_bytes_needed = diskspace.GetBytesFromString(settings.getNeededString(), 0)
            num_suppliers = -1
            if self.customer_idurl == my_id.getIDURL():
                num_suppliers = settings.getSuppliersNumberDesired()
            else:
                known_ecc_map = contactsdb.get_customer_meta_info(self.customer_idurl).get('ecc_map')
                if known_ecc_map:
                    num_suppliers = eccmap.GetEccMapSuppliersNumber(known_ecc_map)
            if num_suppliers > 0:
                self.needed_bytes = int(math.ceil(2.0*total_bytes_needed/float(num_suppliers)))
            else:
                raise Exception('not possible to determine needed_bytes value to be requested from that supplier')
                # self.needed_bytes = int(math.ceil(2.0 * settings.MinimumNeededBytes() / float(settings.DefaultDesiredSuppliers())))

    def A(self, event, *args, **kwargs):
        #---NO_SERVICE---
        if self.state == 'NO_SERVICE':
            if event == 'connect':
                self.state = 'REQUEST'
                self.GoDisconnect = False
                self.doPingRequestService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect':
                self.doReportNoService(*args, **kwargs)
            elif event == 'ack' and self.isServiceAccepted(*args, **kwargs):
                self.state = 'CONNECTED'
                self.doReportConnect(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'disconnect':
                self.state = 'REFUSE'
                self.doCancelServiceQueue(*args, **kwargs)
                self.doCancelService(*args, **kwargs)
            elif event == 'fail' or event == 'connect':
                self.state = 'REQUEST'
                self.GoDisconnect = False
                self.doPingRequestService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect':
                self.state = 'REFUSE'
                self.doCancelService(*args, **kwargs)
            elif event == 'connect':
                self.state = 'REQUEST'
                self.GoDisconnect = False
                self.doPingRequestService(*args, **kwargs)
            elif event == 'fail':
                self.state = 'NO_SERVICE'
                self.doReportNoService(*args, **kwargs)
            elif event == 'ack' and self.isServiceAccepted(*args, **kwargs):
                self.state = 'CONNECTED'
                self.doReportConnect(*args, **kwargs)
        #---REQUEST---
        elif self.state == 'REQUEST':
            if event == 'disconnect':
                self.GoDisconnect = True
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'fail' or (event == 'ack' and not self.isServiceAccepted(*args, **kwargs) and not self.GoDisconnect):
                self.state = 'NO_SERVICE'
                self.doReportNoService(*args, **kwargs)
            elif event == 'ack' and not self.GoDisconnect and self.isServiceAccepted(*args, **kwargs):
                self.state = 'QUEUE?'
                self.doRequestQueueService(*args, **kwargs)
            elif self.GoDisconnect and event == 'ack' and self.isServiceAccepted(*args, **kwargs):
                self.state = 'REFUSE'
                self.doCancelService(*args, **kwargs)
            elif event == 'timer-30sec':
                self.state = 'DISCONNECTED'
                self.doCleanRequest(*args, **kwargs)
                self.doReportDisconnect(*args, **kwargs)
        #---REFUSE---
        elif self.state == 'REFUSE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doCleanRequest(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-10sec' or event == 'fail' or (event == 'ack' and self.isServiceCancelled(*args, **kwargs)):
                self.state = 'NO_SERVICE'
                self.doCleanRequest(*args, **kwargs)
                self.doReportNoService(*args, **kwargs)
        #---QUEUE?---
        elif self.state == 'QUEUE?':
            if event == 'disconnect':
                self.GoDisconnect = True
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif self.GoDisconnect and (event == 'queue-ack' or event == 'queue-fail' or event == 'queue-skip' or event == 'timer-10sec'):
                self.state = 'REFUSE'
                self.doCancelServiceQueue(*args, **kwargs)
                self.doCancelService(*args, **kwargs)
            elif not self.GoDisconnect and (event == 'queue-ack' or event == 'queue-fail' or event == 'queue-skip' or event == 'timer-10sec'):
                self.state = 'CONNECTED'
                self.doReportConnect(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isServiceAccepted(self, *args, **kwargs):
        """
        Condition method.
        """
        newpacket = args[0]
        if strng.to_text(newpacket.Payload).startswith('accepted'):
            if _Debug:
                lg.dbg(_DebugLevel, 'supplier %s accepted my request and will be connected' % self.supplier_idurl)
            return True
        if _Debug:
            lg.dbg(_DebugLevel, 'supplier %s refused my request' % self.supplier_idurl)
        return False

    def isServiceCancelled(self, *args, **kwargs):
        """
        Condition method.
        """
        newpacket = args[0]
        if newpacket.Command == commands.Ack():
            if strng.to_text(newpacket.Payload).startswith('accepted'):
                if _Debug:
                    lg.out(_DebugLevel, 'supplier_connector.isServiceCancelled !!! supplier %s disconnected' % self.supplier_idurl)
                return True
        return False

    def doPingRequestService(self, *args, **kwargs):
        """
        Action method.
        """
        ecc_map = kwargs.get('ecc_map')
        family_position = kwargs.get('family_position')
        family_snapshot = kwargs.get('family_snapshot')
        d = online_status.ping(
            idurl=self.supplier_idurl,
            channel='supplier_connector',
            keep_alive=True,
            ping_retries=3,
        )
        d.addCallback(lambda ok: self._do_request_supplier_service(
            ecc_map=ecc_map,
            family_position=family_position,
            family_snapshot=family_snapshot,
        ))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='supplier_connector.doPingRequestService')
        d.addErrback(lambda err: self.automat('fail', err))

    def doCancelService(self, *args, **kwargs):
        """
        Action method.
        """
        service_info = {}
        # TODO: re-think again about the customer key, do we really need it?
        # my_customer_key_id = my_id.getGlobalID(key_alias='customer')
        # if my_keys.is_key_registered(my_customer_key_id):
        #     service_info['customer_public_key'] = my_keys.get_key_info(
        #         key_id=my_customer_key_id,
        #         include_private=False,
        #         include_signature=False,
        #         generate_signature=False,
        #     )
        ecc_map = kwargs.get('ecc_map')
        if ecc_map:
            service_info['ecc_map'] = ecc_map
        request = p2p_service.SendCancelService(
            remote_idurl=self.supplier_idurl,
            service_name='service_supplier',
            json_payload=service_info,
            callbacks={
                commands.Ack(): self._supplier_service_acked,
                commands.Fail(): self._supplier_service_failed,
            },
        )
        self.request_packet_id = request.PacketID

    def doRequestQueueService(self, *args, **kwargs):
        """
        Action method.
        """
        if not self.queue_subscribe:
            reactor.callLater(0, self.automat, 'queue-skip')  # @UndefinedVariable
            return
        queue_id = global_id.MakeGlobalQueueID(
            queue_alias='supplier-file-modified',
            owner_id=self.customer_id,
            supplier_id=self.supplier_id,
        )
        service_info = {
            'items': [
                {
                    'scope': 'consumer',
                    'action': 'start',
                    'consumer_id': self.customer_id,
                },
                {
                    'scope': 'consumer',
                    'action': 'add_callback',
                    'consumer_id': self.customer_id,
                    'method': self.customer_idurl,
                    'queues': [
                        'supplier-file-modified',
                    ],
                },
                {
                    'scope': 'consumer',
                    'action': 'subscribe',
                    'consumer_id': self.customer_id,
                    'queue_id': queue_id,
                },
            ],
        }
        request = p2p_service.SendRequestService(
            remote_idurl=self.supplier_idurl,
            service_name='service_p2p_notifications',
            json_payload=service_info,
            callbacks={
                commands.Ack(): self._supplier_queue_acked,
                commands.Fail(): self._supplier_queue_failed,
            },
        )
        self.request_queue_packet_id = request.PacketID

    def doCancelServiceQueue(self, *args, **kwargs):
        """
        Action method.
        """
        service_info = {
            'items': [
                {
                    'scope': 'consumer',
                    'action': 'unsubscribe',
                    'consumer_id': self.customer_id,
                    'queue_id': global_id.MakeGlobalQueueID(
                        queue_alias='supplier-file-modified',
                        owner_id=self.customer_id,
                        supplier_id=self.supplier_id,
                    ),
                },
                {
                    'scope': 'consumer',
                    'action': 'remove_callback',
                    'consumer_id': self.customer_id,
                    'method': self.customer_idurl,
                },
                {
                    'scope': 'consumer',
                    'action': 'stop',
                    'consumer_id': self.customer_id,
                },
            ],
        }
        p2p_service.SendCancelService(
            remote_idurl=self.supplier_idurl,
            service_name='service_p2p_notifications',
            json_payload=service_info,
            callbacks={
                commands.Ack(): self._supplier_queue_cancelled,
                commands.Fail(): self._supplier_queue_failed,
            },
        )

    def doCleanRequest(self, *args, **kwargs):
        """
        Action method.
        """
        self.request_packet_id = None
        self.request_queue_packet_id = None

    def doReportNoService(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'supplier_connector.doReportNoService : %s' % self.supplier_idurl)
        for cb_list in list(self.callbacks.values()):
            for cb in cb_list:
                cb(self.supplier_idurl, 'NO_SERVICE')

    def doReportDisconnect(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'supplier_connector.doReportDisconnect : %s' % self.supplier_idurl)
        for cb_list in list(self.callbacks.values()):
            for cb in cb_list:
                cb(self.supplier_idurl, 'DISCONNECTED')

    def doReportConnect(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'supplier_connector.doReportConnect : %s' % self.supplier_idurl)
        for cb_list in list(self.callbacks.values()):
            for cb in cb_list:
                cb(
                    self.supplier_idurl,
                    'CONNECTED',
                    family_position=self._last_known_family_position,
                    ecc_map=self._last_known_ecc_map,
                    family_snapshot=self._last_known_family_snapshot,
                )

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        online_status.remove_online_status_listener_callback(
            idurl=self.supplier_idurl,
            callback_method=self._on_online_status_state_changed,
        )
        connectors(self.customer_idurl).pop(self.supplier_idurl)
        self.latest_supplier_ack = None
        self.request_packet_id = None
        self.request_queue_packet_id = None
        self.supplier_idurl = None
        self.customer_idurl = None
        self.queue_subscribe = None
        self.destroy()

    def _supplier_service_acked(self, response, info):
        if _Debug:
            lg.args(_DebugLevel, response=response, info=info)
        if not self.request_packet_id:
            lg.warn('received "old" response : %r' % response)
            return
        if response.PacketID != self.request_packet_id:
            lg.warn('received "unexpected" response : %r' % response)
            return
        self.latest_supplier_ack = response
        if driver.is_on('service_customer_contracts'):
            the_contract = None
            try:
                if strng.to_text(response.Payload).startswith('accepted:{'):
                    the_contract = jsn.loads_text(strng.to_text(response.Payload)[9:])
            except:
                lg.exc()
            if _Debug:
                lg.args(_DebugLevel, response=response, info=info, contract=the_contract)
            if the_contract:
                if not accounting.verify_storage_contract(the_contract):
                    lg.err('received storage contract from %r is not valid' % self.supplier_idurl)
                    self.latest_supplier_ack = None
                    self.automat('fail', None)
                    return
                from bitdust.customer import payment
                payment.save_storage_contract(self.supplier_idurl, the_contract)
                self.storage_contract = the_contract
        self.automat('ack', response)

    def _supplier_service_failed(self, response, info):
        if _Debug:
            lg.args(_DebugLevel, response=response, info=info)
        self.automat('fail', response or None)

    def _supplier_queue_acked(self, response, info):
        if not self.request_queue_packet_id:
            lg.warn('received "old" queue response : %r' % response)
            return
        if response.PacketID != self.request_queue_packet_id:
            lg.warn('received "unexpected" queue response : %r' % response)
            return
        if _Debug:
            lg.args(_DebugLevel, response=response, info=info)
        # start_consumer(self.customer_id, self.supplier_id)
        self.automat('queue-ack', response)

    def _supplier_queue_cancelled(self, response, info):
        # stop_consumer(self.customer_id, self.supplier_id)
        self.automat('queue-stopped', response)

    def _supplier_queue_failed(self, response, info):
        if _Debug:
            lg.args(_DebugLevel, response=response, info=info)
        self.automat('queue-fail', response or None)

    def _on_online_status_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if oldstate != newstate and newstate in ['CONNECTED', 'OFFLINE']:
            if not (oldstate == 'PING?' and newstate == 'OFFLINE'):
                if _Debug:
                    lg.out(_DebugLevel, 'supplier_connector._on_online_status_state_changed %s : %s->%s, reconnecting now' % (self.supplier_idurl, oldstate, newstate))
                reactor.callLater(0, self.automat, 'connect')  # @UndefinedVariable

    def _do_request_supplier_service(self, ecc_map, family_position, family_snapshot):
        if _Debug:
            lg.args(_DebugLevel, supplier_idurl=self.supplier_idurl, ecc_map=ecc_map, family_position=family_position, family_snapshot=family_snapshot)
        if not self.supplier_idurl:
            lg.warn('supplier idurl is empty, SKIP sending supplier_service request')
            return
        service_info = {
            'customer_id': self.customer_id,
            'needed_bytes': self.needed_bytes,
            # 'minimum_duration_hours': 6,
            # 'maximum_duration_hours': 24*30,
        }
        # TODO: re-think again about the customer key, do we really need it?
        # my_customer_key_id = my_id.getGlobalID(key_alias='customer')
        # if my_keys.is_key_registered(my_customer_key_id):
        #     service_info['customer_public_key'] = my_keys.get_key_info(
        #         key_id=my_customer_key_id,
        #         include_private=False,
        #         include_signature=True,
        #         generate_signature=True,
        #     )
        # else:
        #     lg.warn('my own customer key is not registered: %r' % my_customer_key_id)
        if self.key_id:
            service_info['key_id'] = self.key_id
        self._last_known_ecc_map = ecc_map
        if self._last_known_ecc_map is not None:
            service_info['ecc_map'] = self._last_known_ecc_map
        self._last_known_family_position = family_position
        if self._last_known_family_position is not None:
            service_info['position'] = self._last_known_family_position
        self._last_known_family_snapshot = family_snapshot
        if self._last_known_family_snapshot is not None:
            service_info['family_snapshot'] = id_url.to_bin_list(self._last_known_family_snapshot)
        request = p2p_service.SendRequestService(
            remote_idurl=self.supplier_idurl,
            service_name='service_supplier',
            json_payload=service_info,
            callbacks={
                commands.Ack(): self._supplier_service_acked,
                commands.Fail(): self._supplier_service_failed,
            },
        )
        self.request_packet_id = request.PacketID



#!/usr/bin/python
# service_personal_messages.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_personal_messages.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_personal_messages
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return PersonalMessagesService()


class PersonalMessagesService(LocalService):

    service_name = 'service_personal_messages'
    config_path = 'services/personal-messages/enabled'

    personal_group_key_id = None

    def dependent_on(self):
        return [
            'service_private_groups',
        ]

    def installed(self):
        # TODO: needs more work, service_private_groups() must be reliable first
        return False

    def start(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from twisted.internet.defer import Deferred  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.main import settings
        from bitdust.access import groups
        from bitdust.crypt import my_keys
        from bitdust.userid import my_id
        self.starting_deferred = Deferred()
        self.starting_deferred.addErrback(lambda err: lg.warn('service %r was not started: %r' % (self.service_name, err.getErrorMessage() if err else 'unknown reason')))
        self.personal_group_key_id = my_keys.make_key_id(alias='person', creator_glob_id=my_id.getGlobalID())

        # TODO: needs more work, service_private_groups() must be reliable first
        return True

        if not my_keys.is_key_registered(self.personal_group_key_id):
            self.personal_group_key_id = groups.create_new_group(
                label='my personal messages',
                creator_id=my_id.getGlobalID(),
                key_size=settings.getPrivateKeySize(),
                group_alias='person',
                with_group_info=True,
            )
            if self.personal_group_key_id is None:
                lg.err('failed creating "personal" message group, service_personal_messages() can not be started')
                if self.starting_deferred:
                    if not self.starting_deferred.called:
                        self.starting_deferred.errback(Exception('failed creating "personal" message group'))
                return self.starting_deferred
        d = groups.send_group_pub_key_to_suppliers(self.personal_group_key_id)
        d.addCallback(lambda results: self._do_join_my_personal_group())
        d.addErrback(self.starting_deferred.errback)
        d.addTimeout(10, clock=reactor)
        return self.starting_deferred

    def stop(self):
        # from bitdust.interface import api
        # api.group_leave(self.personal_group_key_id, erase_key=False)
        return True

    def _do_join_my_personal_group(self):
        from bitdust.logs import lg
        from bitdust.interface import api
        lg.info('about to join my personal message group: %r' % self.personal_group_key_id)
        api.group_join(self.personal_group_key_id)
        if self.starting_deferred:
            if not self.starting_deferred.called:
                self.starting_deferred.callback(True)
            self.starting_deferred = None



#!/usr/bin/env python
# contract_chain_consumer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (contract_chain_consumer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: contract_chain_consumer
.. role:: red

BitDust contract_chain_consumer() Automat

EVENTS:
    * :red:`accountants-connected`
    * :red:`accountants-failed`
    * :red:`init`
    * :red:`miner-connected`
    * :red:`miner-failed`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`timer-1min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.python.failure import Failure

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.p2p import lookup
from bitdust.p2p import p2p_service
from bitdust.p2p import p2p_service_seeker

#------------------------------------------------------------------------------

_ContractChainConsumer = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _ContractChainConsumer
    if event is None and not args:
        return _ContractChainConsumer
    if _ContractChainConsumer is None:
        # set automat name and starting state here
        _ContractChainConsumer = ContractChainConsumer('contract_chain_consumer', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _ContractChainConsumer.automat(event, *args, **kwargs)
    return _ContractChainConsumer


#------------------------------------------------------------------------------


class ContractChainConsumer(automat.Automat):
    """
    This class implements all the functionality of the ``contract_chain_consumer()`` state machine.
    """

    timers = {
        'timer-1min': (60, ['MINER?', 'ACCOUNTANTS?']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of contract_chain_consumer() machine.
        """
        self.connected_accountants = []
        self.connected_miner = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://bitdust.io/visio2python/>`_ tool.
        """
        #---MINER?---
        if self.state == 'MINER?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'miner-connected':
                self.state = 'CONNECTED'
            elif event == 'stop' or event == 'timer-1min' or event == 'miner-failed':
                self.state = 'DISCONNECTED'
                self.doDisconnectAccountants(*args, **kwargs)
                self.doDisconnectMiner(*args, **kwargs)
        #---AT_STARTUP---
        elif self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'DISCONNECTED'
                self.doInit(*args, **kwargs)
        #---ACCOUNTANTS?---
        elif self.state == 'ACCOUNTANTS?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop' or event == 'timer-1min' or event == 'accountants-failed':
                self.state = 'DISCONNECTED'
                self.doDisconnectAccountants(*args, **kwargs)
            elif event == 'accountants-connected':
                self.state = 'MINER?'
                self.doConnectMiner(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'start':
                self.state = 'ACCOUNTANTS?'
                self.doConnectAccountants(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'DISCONNECTED'
                self.doDisconnectAccountants(*args, **kwargs)
                self.doDisconnectMiner(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """

    def doConnectAccountants(self, *args, **kwargs):
        """
        Action method.
        """
        self.accountant_lookups = 0
        self._lookup_next_accountant()

    def doDisconnectAccountants(self, *args, **kwargs):
        """
        Action method.
        """
        for idurl in self.connected_accountants:
            p2p_service.SendCancelService(idurl, 'service_accountant')
        self.connected_accountants = []

    def doConnectMiner(self, *args, **kwargs):
        """
        Action method.
        """
        self.miner_lookups = 0
        self._lookup_miner()

    def doDisconnectMiner(self, *args, **kwargs):
        """
        Action method.
        """
        if self.connected_miner:
            p2p_service.SendCancelService(self.connected_miner, 'service_miner')
        self.connected_miner = None

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.destroy()
        global _ContractChainConsumer
        del _ContractChainConsumer
        _ContractChainConsumer = None

    #------------------------------------------------------------------------------

    def _on_accountant_lookup_finished(self, idurl):
        if self.state != 'ACCOUNTANTS?':
            lg.warn('internal state was changed during accountant lookup, SKIP next lookup')
            return None
        if not idurl or isinstance(idurl, Exception) or isinstance(idurl, Failure):
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._on_accountant_lookup_finished with no results, try again')
            reactor.callLater(0, self._lookup_next_accountant)  # @UndefinedVariable
            return None
        if idurl in self.connected_accountants:
            lg.warn('node %s already connected as accountant')
        else:
            self.connected_accountants.append(idurl)
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._on_accountant_lookup_finished !!!!!!!! %s CONNECTED as new accountant' % idurl)
        reactor.callLater(0, self._lookup_next_accountant)  # @UndefinedVariable
        return None

    def _lookup_next_accountant(self):
        if len(self.connected_accountants) >= 3:  # TODO: read from settings.: max accountants
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._lookup_next_accountant SUCCESS, %d accountants connected' % len(self.connected_accountants))
            self.automat('accountants-connected')
            return
        if self.accountant_lookups >= 10:  # TODO: read from settings.
            if len(self.connected_accountants) >= 1:  # TODO: read from settings: min accountants
                if _Debug:
                    lg.out(_DebugLevel, 'contract_chain_consumer._lookup_next_accountant FAILED after %d retries, but %d accountants connected' % (self.accountant_lookups, len(self.connected_accountants)))
                self.automat('accountants-connected')
                return
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._lookup_next_accountant FAILED after %d retries with no results' % self.accountant_lookups)
            self.automat('accountants-failed')
            return
        self.accountant_lookups += 1
        p2p_service_seeker.connect_random_node(
            'service_accountant',
            lookup_method=lookup.random_merchant,
            service_params={
                'action': 'read',
            },
            exclude_nodes=self.connected_accountants,
        ).addBoth(self._on_accountant_lookup_finished)

    def _on_miner_lookup_finished(self, idurl):
        if not idurl or isinstance(idurl, Exception) or isinstance(idurl, Failure):
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._on_miner_lookup_finished with no results, try again')
            reactor.callLater(0, self._lookup_miner)  # @UndefinedVariable
            return None
        self.connected_miner = idurl
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_consumer._on_miner_lookup_finished SUCCESS, miner %s connected' % self.connected_miner)
        self.automat('miner-connected')

    def _lookup_miner(self):
        if self.miner_lookups >= 5:  # TODO: read value from settings
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_consumer._lookup_miner FAILED after %d retries' % self.miner_lookups)
            self.automat('miner-failed')
            return
        self.miner_lookups += 1
        p2p_service_seeker.connect_random_node(
            'service_miner',
            lookup_method=lookup.random_merchant,  # TODO: rework if needed
        ).addBoth(self._on_miner_lookup_finished)



#!/usr/bin/python
# service_my_data.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_my_data.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_my_data
"""

from __future__ import absolute_import
from bitdust.services.local_service import SlowStartingLocalService


def create_service():
    return MyDataService()


class MyDataService(SlowStartingLocalService):

    service_name = 'service_my_data'
    config_path = 'services/my-data/enabled'

    last_time_keys_synchronized = None

    def dependent_on(self):
        return [
            'service_keys_storage',
        ]

    def init(self, **kwargs):
        SlowStartingLocalService.init(self, **kwargs)
        from bitdust.main import events
        events.add_subscriber(self._on_my_storage_ready, 'my-storage-ready')
        events.add_subscriber(self._on_my_storage_not_ready_yet, 'my-storage-not-ready-yet')

    def shutdown(self):
        from bitdust.main import events
        events.remove_subscriber(self._on_my_storage_not_ready_yet, 'my-storage-not-ready-yet')
        events.remove_subscriber(self._on_my_storage_ready, 'my-storage-ready')
        SlowStartingLocalService.shutdown(self)

    def start(self):
        from bitdust.logs import lg
        from bitdust.main import listeners
        from bitdust.storage import keys_synchronizer
        from bitdust.storage import index_synchronizer
        from bitdust.storage import backup_fs
        if keys_synchronizer.is_synchronized() and index_synchronizer.is_synchronized():
            self.confirm_service_started(result=True)
            if listeners.is_populate_required('private_file'):
                # listeners.populate_later().remove('private_file')
                backup_fs.populate_private_files()
        else:
            lg.warn('can not start service_my_data right now, keys_synchronizer.is_synchronized=%r index_synchronizer.is_synchronized=%r' % (keys_synchronizer.is_synchronized(), index_synchronizer.is_synchronized()))
        return self.starting_deferred

    def stop(self):
        return True

    def health_check(self):
        return True

    def _on_my_storage_ready(self, evt):
        from bitdust.logs import lg
        from bitdust.main import listeners
        from bitdust.services import driver
        from bitdust.storage import backup_fs
        if self.starting_deferred:
            self.confirm_service_started(result=True)
            if listeners.is_populate_required('private_file'):
                # listeners.populate_later().remove('private_file')
                backup_fs.populate_private_files()
        if driver.is_enabled('service_my_data'):
            if not driver.is_started('service_my_data'):
                lg.info('my storage is ready, starting service_my_data()')
                driver.start_single('service_my_data')

    def _on_my_storage_not_ready_yet(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if self.starting_deferred:
            self.confirm_service_started(result=Exception('my storage is not ready yet'))
        if driver.is_enabled('service_my_data'):
            if not driver.is_started('service_my_data'):
                lg.info('my storage is not ready yet, stopping service_my_data()')
                driver.stop_single('service_my_data')



"""
Peers handler module for Bismuth nodes
@EggPoolNet
"""

import json
import os
import random
import shutil
# import re
import sys
import threading
from time import time

import socks

import connections
import regnet
from essentials import most_common_dict, percentage_in

__version__ = '0.0.19'

_LogStatusMessages = False


class Peers:
    """The peers manager. A thread safe peers manager"""

    __slots__ = (
        'app_log',
        'config',
        'logstats',
        'node',
        'peersync_lock',
        'startup_time',
        'reset_time',
        'warning_list',
        'stats',
        'connection_pool',
        'peer_opinion_dict',
        'consensus_percentage',
        'consensus',
        'tried',
        'peer_dict',
        'peerfile',
        'suggested_peerfile',
        'banlist',
        'whitelist',
        'ban_threshold',
        'ip_to_mainnet',
        'peers',
        'accept_peers',
        'peerlist_updated',
    )

    def __init__(self, app_log, config=None, logstats=True, node=None):
        self.app_log = app_log
        self.config = config
        self.logstats = logstats
        self.peersync_lock = threading.Lock()
        self.startup_time = time()
        self.reset_time = self.startup_time
        self.warning_list = []
        self.stats = []
        self.peer_opinion_dict = {}
        self.consensus_percentage = 0
        self.consensus = None
        self.tried = {}
        self.peer_dict = {}
        self.ip_to_mainnet = {}
        self.connection_pool = []
        # We store them apart from the initial config, could diverge somehow later on.
        self.banlist = config.banlist
        self.whitelist = config.whitelist
        self.ban_threshold = config.ban_threshold
        self.accept_peers = config.accept_peers

        self.peerfile = 'peers.txt'
        self.suggested_peerfile = 'suggested_peers.txt'
        self.peerlist_updated = False

        self.node = node

        if self.is_testnet:  # overwrite for testnet
            self.peerfile = 'peers_test.txt'
            self.suggested_peerfile = 'suggested_peers_test.txt'

        if self.is_regnet:  # regnet won't use any peer, won't connect. Kept for compatibility
            self.peerfile = regnet.REGNET_PEERS
            self.suggested_peerfile = regnet.REGNET_SUGGESTED_PEERS

    @property
    def is_testnet(self):
        """Helper to check if testnet or not. Only one place to change variable names and test"""
        if self.config.regnet:
            # regnet takes over testnet
            return False
        if self.config.testnet:
            return True
        return 'testnet' in self.config.version

    @property
    def is_regnet(self):
        """Helper to check if regnet or not. Only one place to change variable names and test"""
        if self.config.regnet:
            # regnet takes over testnet
            return True
        return 'regnet' in self.config.version

    def dict_shuffle(self, dictinary):
        l = list(dictinary.items())
        random.shuffle(l)
        return dict(l)

    def status_dict(self):
        """Returns a status as a dict"""
        status = {'version': self.config.VERSION, 'stats': self.stats}
        return status

    def store_mainnet(self, ip, version):
        """Stores the mainnet version of a peer. Can't change unless reconnects"""
        self.ip_to_mainnet[ip] = version

    def forget_mainnet(self, ip):
        """Peers disconnected, forget his mainnet version"""
        self.ip_to_mainnet.pop(ip, None)

    def version_allowed(self, ip, version_allow):
        """
        If we don't know the version for this ip, allow.
        If we know, check
        """
        if ip not in self.ip_to_mainnet:
            return True
        return self.ip_to_mainnet[ip] in version_allow

    def peers_test(self, file, peerdict: dict, strict=True):
        """Validates then adds a peer to the peer list on disk"""
        # called by Sync, should not be an issue, but check if needs to be thread safe or not.
        # also called by self.client_loop, which is to be reworked
        # Egg: Needs to be thread safe.
        self.peerlist_updated = False
        try:
            with open(file, 'r') as peer_file:
                peers_pairs = json.load(peer_file)
            # print('Peers file test', file, len(peers_pairs), len(peerdict))
            # TODO: rework, because this takes too much time and freezes the status thread.
            # to be done in a dedicated thread, with one peer per xx seconds, not all at once, and added properties.
            for ip, port in dict(peerdict).items():
                # I do create a new dict copy above, because logs showed that the dict can change while iterating
                if self.node.IS_STOPPING:
                    # Early exit if stopping
                    return
                try:
                    if ip not in peers_pairs:
                        self.app_log.debug(f'Testing connectivity to: {ip}:{port}')
                        s = socks.socksocket()
                        try:
                            # connect timeout
                            s.settimeout(5)
                            if self.config.tor:
                                s.setproxy(socks.PROXY_TYPE_SOCKS5, '127.0.0.1', 9050)
                            if strict:
                                s.connect((ip, int(port)))
                                connections.send(s, 'getversion')
                                versiongot = connections.receive(s, timeout=1)
                                if versiongot == '*':
                                    raise ValueError('peer busy')
                                if versiongot not in self.config.version_allow:
                                    raise ValueError(f'cannot save {ip}, incompatible protocol version {versiongot} '
                                                     f'not in {self.config.version_allow}', )
                                self.app_log.debug(f'Inbound: Distant peer {ip}:{port} responding: {versiongot}')
                            else:
                                s.connect((ip, int(port)))
                        finally:
                            # properly end the connection in all cases
                            try:
                                s.close()
                            except:
                                pass
                        peers_pairs[ip] = port
                        self.app_log.debug(f'Inbound: Peer {ip}:{port} saved to peers')
                        self.peerlist_updated = True
                    else:
                        self.app_log.debug('Distant peer already in peers')

                except Exception as e:
                    # exception for a single peer
                    self.app_log.debug(f'Inbound: Distant peer not connectible ({e})')

            if self.peerlist_updated:
                self.app_log.debug(f'{file} peerlist updated ({len(peers_pairs)}) total')  # the whole dict is saved
                with open(f'{file}.tmp', 'w') as peer_file:
                    json.dump(peers_pairs, peer_file)
                shutil.move(f'{file}.tmp', file)
            else:
                self.app_log.debug(f'{file} peerlist update skipped, no changes')

        except Exception as e:
            # Exception for the file itself.
            self.app_log.warning(f"Error reading {file}: '{e}'")

    def append_client(self, client):
        """
        :param client: a string "ip:port"
        :return:
        """
        # TODO: thread safe?
        self.connection_pool.append(client)
        self.del_try(client)

    def remove_client(self, client):
        # TODO: thread safe?
        if client in self.connection_pool:
            try:
                self.app_log.debug(f'Will remove {client} from active pool')
                self.connection_pool.remove(client)
            except:
                raise

    def unban(self, peer_ip):
        """Removes the peer_ip from the warning list"""
        # TODO: Not thread safe atm. Should use a thread aware list or some lock
        if peer_ip in self.warning_list:
            self.warning_list.remove(peer_ip)
            self.app_log.warning(f'Removed a warning for {peer_ip}')

    def warning(self, sdef, ip, reason, count):
        """Adds a weighted warning to a peer."""
        # TODO: Not thread safe atm. Should use a thread aware list or some lock
        if ip not in self.whitelist:
            # TODO: use a dict instead of several occurrences in a list
            for x in range(count):
                self.warning_list.append(ip)
            self.app_log.warning(f'Added {count} warning(s) to {ip}: {reason} '
                                 f'({self.warning_list.count(ip)} / {self.ban_threshold})', )
            if self.warning_list.count(ip) >= self.ban_threshold:
                self.banlist.append(ip)
                self.app_log.warning(f'{ip} is banned: {reason}')
                return True
            else:
                return False

    def peers_get(self, peer_file=''):
        """Returns a peer_file from disk as a dict {ip:port}"""
        peer_dict = {}
        try:
            if not peer_file:
                peer_file = self.peerfile
            if not os.path.exists(peer_file):
                with open(peer_file, 'w') as fp:
                    # was "a": append would risk adding stuff to a file create in the mean time.
                    self.app_log.warning(f'Peer file created in {peer_file}')
                    fp.write('{}')  # empty dict. An empty string is not json valid.
            else:
                with open(peer_file, 'r') as fp:
                    peer_dict = json.load(fp)
        except Exception as e:
            self.app_log.warning(f'Error peers_get {e} reading {peer_file}')
        return peer_dict

    def peer_list_disk_format(self):
        """Returns a peerfile as is, simple text format or json, as it is on disk"""
        # TODO: caching and format to handle here
        with open(self.peerfile, 'r') as peer_list:
            peers = peer_list.read()
        return peers

    @property
    def consensus_most_common(self):
        """Consensus vote"""
        try:
            return most_common_dict(self.peer_opinion_dict)
        except:
            # no consensus yet
            return 0

    @property
    def consensus_max(self):
        try:
            return max(self.peer_opinion_dict.values())
        except:
            # no consensus yet
            return 0

    @property
    def consensus_size(self):
        """Number of nodes in consensus"""
        return len(self.peer_opinion_dict)

    def is_allowed(self, peer_ip, command=''):
        """Tells if the given peer is allowed for that command"""
        # TODO: more granularity here later
        # Always allow whitelisted ip to post as block
        if 'block' == command and self.is_whitelisted(peer_ip):
            return True
        # always allowed commands, only required and non cpu intensive.
        if command in ('portget', ):
            return True
        # only allow local host for "stop" and addpeers command
        if command in ('stop', 'addpeers'):
            return peer_ip == '127.0.0.1'
        return peer_ip in self.config.allowed or 'any' in self.config.allowed

    def is_whitelisted(self, peer_ip, command=''):
        # TODO: could be handled later on via "allowed" and rights.
        return peer_ip in self.whitelist or '127.0.0.1' == peer_ip

    def is_banned(self, peer_ip) -> bool:
        return peer_ip in self.banlist

    def dict_validate(self, json_dict: str) -> str:
        """temporary fix for broken peerlists"""
        if json_dict.count('}') > 1:
            result = json_dict.split('}')[0] + '}'
        else:
            result = json_dict
        return result

    def peersync(self, subdata: str) -> int:
        """Got a peers list from a peer, process. From worker().
        returns the number of added peers, -1 if it was locked or not accepting new peers
        subdata is a dict, { 'ip': 'port'}"""
        # early exit to reduce future levels
        if not self.config.accept_peers:
            return -1
        if self.peersync_lock.locked():
            # TODO: means we will lose those peers forever.
            # TODO: buffer, and keep track of recently tested peers.
            self.app_log.debug('Outbound: Peer sync occupied')
            return -1
        # Temp fix: subdata is typed str, but we have a dict sometimes.
        if type(subdata) == dict:
            # Enforce expected type.
            self.app_log.warning('Enforced expected type for peersync subdata')
            subdata = json.dumps(subdata)
        with self.peersync_lock:
            try:
                total_added = 0

                subdata = self.dict_validate(subdata)
                data_dict = json.loads(subdata)

                self.app_log.debug(f'Received {len(data_dict)} peers.')
                # Simplified the log, every peers then has a ok or ko status anyway.
                for ip, port in data_dict.items():
                    if ip not in self.peer_dict:
                        self.app_log.debug(f'Outbound: {ip}:{port} is a new peer, saving if connectible')
                        try:
                            s_purge = socks.socksocket()
                            s_purge.settimeout(5)
                            if self.config.tor:
                                s_purge.setproxy(socks.PROXY_TYPE_SOCKS5, '127.0.0.1', 9050)
                            s_purge.connect((ip, int(port)))  # save a new peer file with only active nodes
                            s_purge.close()
                            # This only adds to our local dict, does not force save.
                            if ip not in self.peer_dict:
                                total_added += 1
                                self.peer_dict[ip] = port
                                self.app_log.debug(f'Inbound: Peer {ip}:{port} saved to local peers')
                        except:
                            self.app_log.debug('Not connectible')
                    else:
                        self.app_log.debug(f'Outbound: {ip}:{port} is not a new peer')
            except Exception as e:
                self.app_log.warning(e)
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                print(exc_type, fname, exc_tb.tb_lineno)
                raise
        return total_added

    def consensus_add(self, peer_ip, consensus_blockheight, sdef, last_block):
        # obviously too old blocks, we have half a day worth of validated blocks after them
        # no ban, they can (should) be syncing but they can't possibly be in consensus list.
        too_old = last_block - 720
        try:
            if peer_ip not in self.peer_opinion_dict:
                if consensus_blockheight < too_old:
                    self.app_log.warning(f'{peer_ip} received block too old ({consensus_blockheight}) for consensus')
                    return

            self.app_log.debug(f'Updating {peer_ip} in consensus')
            self.peer_opinion_dict[peer_ip] = consensus_blockheight

            self.consensus = most_common_dict(self.peer_opinion_dict)

            self.consensus_percentage = percentage_in(self.peer_opinion_dict[peer_ip], self.peer_opinion_dict.values())

            if int(consensus_blockheight) > int(self.consensus) + 30 and self.consensus_percentage > 50 and len(self.peer_opinion_dict) > 10:
                if self.warning(sdef, peer_ip, f'Consensus deviation too high, {peer_ip} banned', 10):
                    return

        except Exception as e:
            self.app_log.warning(e)
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            raise

    def consensus_remove(self, peer_ip):
        if peer_ip in self.peer_opinion_dict:
            try:
                self.app_log.debug(f'Will remove {peer_ip} from consensus pool {self.peer_opinion_dict}')
                self.peer_opinion_dict.pop(peer_ip, None)
            except:
                raise

    def can_connect_to(self, host, port):
        """
        Tells if we can connect to this host
        :param host:
        :param port:
        :return:
        """
        if host in self.banlist:
            return False  # Banned IP
        host_port = f'{host}:{port}'
        if host_port in self.connection_pool:
            return False  # Already connected to
        try:
            tries, timeout = self.tried[host_port]
        except:
            tries, timeout = 0, 0  # unknown host for now, never tried.
        if timeout > time():
            return False  # We tried before, timeout is not expired.
        if self.is_whitelisted(host):
            return True  # whitelisted peers are always connectible, without variability condition.
        # variability test.
        c_class = '.'.join(host.split('.')[:-1]) + '.'
        matching = [ip_port for ip_port in self.connection_pool if c_class in ip_port]
        # If we already have 2 peers from that C ip class in our connection pool, ignore.
        if len(matching) >= 2:
            # Temp debug
            self.app_log.warning(f'Ignoring {host_port} since we already have 2 ips of that C Class in our pool.')
            return False
        # Else we can
        return True

    def add_try(self, host, port):
        """
        Add the host to the tried dict with matching timeout depending on its state.
        :param host:
        :param port:
        :return:
        """
        host_port = f'{host}:{port}'
        try:
            tries, timeout = self.tried[host_port]
        except:
            tries, timeout = 0, 0
        if tries <= 0:  # First time can be temp, retry again
            delay = 30
        elif tries == 1:  # second time, give it 5 minutes
            delay = 5*60
        elif tries == 2:  # third time, give it 15 minutes
            delay = 15*60
        else:  # 30 minutes before trying again
            delay = 30*60
        tries += 1
        if tries > 3:
            tries = 3
        self.tried[host_port] = (tries, time() + delay)
        # Temp
        self.app_log.debug(f'Set timeout {delay} try {tries} for {host_port}')

    def del_try(self, host, port=None):
        """
        Remove the peer from tried list. To be called when we successfully connected.
        :param host: an ip as a string, or an "ip:port" string
        :param port: optional, port as an int
        :return:
        """
        try:
            if port:
                host_port = f'{host}:{port}'
            else:
                host_port = host
            if host_port in self.tried:
                del self.tried[host_port]
        except Exception as e:
            print(e)
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)

    def reset_tried(self):
        """
        Remove the older timeouts from the tried list.
        Keep the recent ones or we end up trying the first ones again and again
        """
        limit = time() + 12*60  # matches 2.5 tries :)
        remove = [client for client in self.tried if self.tried[client][1] > limit]
        for client in remove:
            del self.tried[client]

    def client_loop(self, node, this_target):
        """Manager loop called every 30 sec. Handles maintenance"""
        try:
            for key, value in dict(self.dict_shuffle(self.peer_dict)).items():
                # The dict() above is not an error or a cast,
                # it's to make a copy of the dict and avoid "dictionary changed size during iteration"
                host = key
                port = int(value)
                if host == self.node.host:
                    self.app_log.debug(f'Skip connecting to {host}:{port}')
                    continue

                if self.is_testnet:
                    port = 2829
                if threading.active_count()/3 < self.config.thread_limit and self.can_connect_to(host, port):
                    self.app_log.debug(f'Will attempt to connect to {host}:{port}')
                    self.add_try(host, port)
                    t = threading.Thread(target=this_target, args=(host, port, node), name=f'out_{host}_{port}')  # threaded connectivity to nodes here
                    self.app_log.debug(f'---Starting a client thread {threading.currentThread()} ---')
                    t.daemon = True
                    t.start()

            if len(self.peer_dict) < 6 and int(time() - self.startup_time) > 30:
                # join in random peers after x seconds
                self.app_log.debug('Not enough peers in consensus, joining in peers suggested by other nodes')
                self.peer_dict.update(self.peers_get(self.suggested_peerfile))

            if len(self.connection_pool) < self.config.nodes_ban_reset and int(time() - self.startup_time) > 15:
                # do not reset before 30 secs have passed
                self.app_log.debug(f'Only {len(self.connection_pool)} connections active, resetting banlist')
                del self.banlist[:]
                self.banlist.extend(self.config.banlist)  # reset to config version
                del self.warning_list[:]

            if len(self.connection_pool) < 10:
                self.app_log.debug(f'Only {len(self.connection_pool)} connections active, '
                                     f'resetting the connection history', )
                # TODO: only reset large timeouts, or we end up trying the sames over and over if we never get to 10.
                # self.
                self.reset_tried()

            if self.config.nodes_ban_reset <= len(self.banlist) and len(self.connection_pool) <= len(self.banlist) \
                    and (time() - self.reset_time) > 60 * 10:
                # do not reset too often. 10 minutes here
                self.app_log.debug(f'Less active connections ({len(self.connection_pool)}) '
                                     f'than banlist ({len(self.banlist)}), resetting banlist and tried list', )
                del self.banlist[:]
                self.banlist.extend(self.config.banlist)  # reset to config version
                del self.warning_list[:]
                self.reset_tried()
                self.reset_time = time()

            # if _LogStatusMessages:
            #     self.app_log.debug('Status: Testing peers')
            self.peer_dict.update(self.peers_get(self.peerfile))
            # self.peer_dict.update(self.peers_get(self.suggested_peerfile))

            # TODO: this is not OK. client_loop is called every 30 sec and should NOT contain any lengthy calls.
            self.peers_test(self.suggested_peerfile, self.peer_dict, strict=False)
            self.peers_test(self.peerfile, self.peer_dict, strict=True)

        except Exception as e:
            self.app_log.warning(f'Status: peers client loop skipped due to error: {e}')
            # raise
            """We do not want to raise here, since the rest of the calling method would be skipped also.
            It's ok to skip this part only
            The calling method has other important subsequent calls that have to be done.
            """

    def status_log(self):
        """Prints the peers part of the node status"""
        if not _LogStatusMessages:
            return
        if self.banlist:
            self.app_log.warning(f'Status: Banlist: {self.banlist}')
            self.app_log.warning(f'Status: Banlist Count : {len(self.banlist)}')
        if self.whitelist:
            self.app_log.warning(f'Status: Whitelist: {self.whitelist}')

        self.app_log.warning(f'Status: Known Peers: {len(self.peer_dict)}')
        self.app_log.debug(f'Status: Tried: {self.tried}')
        self.app_log.debug(f'Status: Tried Count: {len(self.tried)}')
        self.app_log.debug(f'Status: List of Outbound connections: {self.connection_pool}')
        self.app_log.warning(f'Status: Number of Outbound connections: {len(self.connection_pool)}')
        if self.consensus:  # once the consensus is filled
            self.app_log.warning(f'Status: Consensus height: {self.consensus} = {self.consensus_percentage}%')
            self.app_log.warning(f'Status: Last block opinion: {self.peer_opinion_dict}')
            self.app_log.warning(f'Status: Total number of nodes: {len(self.peer_opinion_dict)}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import functools
from random import choice
from six.moves import range


def cache1lvl(maxsize=100):
    def decorating_function(user_function):
        cache1lvl = {}

        @functools.wraps(user_function)
        def wrapper(key, *args, **kwargs):
            try:
                result = cache1lvl[key]
            except KeyError:
                if len(cache1lvl) == maxsize:
                    for i in range(maxsize // 10 or 1):
                        del cache1lvl[choice(list(cache1lvl.keys()))]
                cache1lvl[key] = user_function(key, *args, **kwargs)
                result = cache1lvl[key]
#                result = user_function(obj, key, *args, **kwargs)
            return result

        def clear():
            cache1lvl.clear()

        def delete(key):
            try:
                del cache1lvl[key]
                return True
            except KeyError:
                return False

        wrapper.clear = clear
        wrapper.cache = cache1lvl
        wrapper.delete = delete
        return wrapper
    return decorating_function


def cache2lvl(maxsize=100):
    def decorating_function(user_function):
        cache = {}

        @functools.wraps(user_function)
        def wrapper(*args, **kwargs):
#            return user_function(*args, **kwargs)
            try:
                result = cache[args[0]][args[1]]
            except KeyError:
#                print wrapper.cache_size
                if wrapper.cache_size == maxsize:
                    to_delete = maxsize // 10 or 1
                    for i in range(to_delete):
                        key1 = choice(list(cache.keys()))
                        key2 = choice(list(cache[key1].keys()))
                        del cache[key1][key2]
                        if not cache[key1]:
                            del cache[key1]
                    wrapper.cache_size -= to_delete
#                print wrapper.cache_size
                result = user_function(*args, **kwargs)
                try:
                    cache[args[0]][args[1]] = result
                except KeyError:
                    cache[args[0]] = {args[1]: result}
                wrapper.cache_size += 1
            return result

        def clear():
            cache.clear()
            wrapper.cache_size = 0

        def delete(key, inner_key=None):
            if inner_key:
                try:
                    del cache[key][inner_key]
                    if not cache[key]:
                        del cache[key]
                    wrapper.cache_size -= 1
                    return True
                except KeyError:
                    return False
            else:
                try:
                    wrapper.cache_size -= len(cache[key])
                    del cache[key]
                    return True
                except KeyError:
                    return False

        wrapper.clear = clear
        wrapper.cache = cache
        wrapper.delete = delete
        wrapper.cache_size = 0
        return wrapper
    return decorating_function



#!/usr/bin/env python
# bpmain.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (bpmain.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: bpmain.

This is the entry point of the program, see method ``main()`` bellow.
"""

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import sys
import time
import threading

#-------------------------------------------------------------------------------

_AppDataDir = None

#-------------------------------------------------------------------------------


def print_text(msg, nl='\n'):
    """
    Send some text output to the console.
    """
    sys.stdout.write(msg + nl)
    sys.stdout.flush()


#-------------------------------------------------------------------------------


def show():
    """
    Just calls ``p2p.web.control.show()`` to open the GUI.
    """
    # from bitdust.main import control
    # TODO: raise up electron window?
    return 0


def init(UI='', options=None, args=None, overDict=None, executablePath=None):
    """
    In the method ``main()`` program firstly checks the command line arguments
    and then calls this method to start the whole process.

    This initialize some low level modules and finally create an
    instance of ``initializer()`` state machine and send it an event
    "run".
    """
    global _AppDataDir

    from bitdust.logs import lg
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.init UI="%s"' % UI)

    from bitdust.system import bpio
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.init ostype=%r' % bpio.ostype())

    #---settings---
    from bitdust.main import initializer
    initializer.init_settings(
        base_dir=_AppDataDir,
        override_configs=overDict,
        enable_debug=(not options or options.debug is None),
    )
    from bitdust.main import settings
    from bitdust.main import config

    #---USE_TRAY_ICON---
    USE_TRAY_ICON = False
    if _Debug:
        lg.out(_DebugLevel, '    USE_TRAY_ICON=' + str(USE_TRAY_ICON))
    if USE_TRAY_ICON:
        from bitdust.system import tray_icon
        icons_path = bpio.portablePath(os.path.join(bpio.getExecutableDir(), 'icons'))
        if _Debug:
            lg.out(_DebugLevel, 'bpmain.init call tray_icon.init(%s)' % icons_path)
        tray_icon.init(icons_path)

        def _tray_control_func(cmd):
            if cmd == 'exit':
                from bitdust.main import shutdowner
                shutdowner.A('stop', 'exit')

        tray_icon.SetControlFunc(_tray_control_func)

    #---OS Windows init---
    if bpio.Windows():
        try:
            from win32event import CreateMutex  # @UnresolvedImport
            mutex = CreateMutex(None, False, 'BitDust')
            if _Debug:
                lg.out(_DebugLevel, 'bpmain.init created a Mutex: %s' % str(mutex))
        except:
            lg.exc()

    #---twisted reactor---
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.init want to import twisted.internet.reactor')
    try:
        from twisted.internet import reactor  # @UnresolvedImport
    except:
        lg.exc()
        sys.exit('Error initializing reactor in bpmain.py\n')
        return

    initializer.init_engine()

    #---memdebug---
    if config.conf().getBool('logs/memdebug-enabled'):
        try:
            from bitdust.logs import memdebug
            memdebug_port = int(config.conf().getData('logs/memdebug-port'))
            memdebug.start(memdebug_port)
            reactor.addSystemEventTrigger('before', 'shutdown', memdebug.stop)  # @UndefinedVariable
            if _Debug:
                lg.out(_DebugLevel, 'bpmain.init memdebug web server started on port %d' % memdebug_port)
        except:
            lg.exc()

    #---process ID---
    try:
        pid = os.getpid()
        pid_file_path = os.path.join(settings.AppDataDir(), 'processid')
        bpio.WriteTextFile(pid_file_path, str(pid))
        if _Debug:
            lg.out(_DebugLevel, 'bpmain.init wrote process id [%s] in the file %s' % (str(pid), pid_file_path))
    except:
        lg.exc()

    #---reactor.callLater patch---
    # if _Debug:
    #     patchReactorCallLater(reactor)
    #     monitorDelayedCalls(reactor)

    if _Debug:
        lg.out(_DebugLevel, '    python executable is: %s' % sys.executable)
        lg.out(_DebugLevel, '    python version is:\n%s' % sys.version)
        lg.out(_DebugLevel, '    python sys.path is:\n                %s' % ('\n                '.join(sys.path)))
        lg.out(_DebugLevel, '\n' + bpio.osinfofull())

    if _Debug:
        lg.out(_DebugLevel, 'bpmain.init going to initialize state machines')

    #---START!---
    initializer.init_automats()

    IA = initializer.A()
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.init is sending event "run" to initializer()')
    if bpio.Android():
        IA.automat('run', UI)
    else:
        reactor.callWhenRunning(IA.automat, 'run', UI)  # @UndefinedVariable
    return IA


#------------------------------------------------------------------------------


def shutdown():
    from bitdust.logs import lg
    # from bitdust.main import config
    # from bitdust.system import bpio
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.shutdown')

    from bitdust.main import shutdowner
    shutdowner.A('reactor-stopped')

    shutdowner.shutdown_automats()

    # from bitdust.main import listeners
    # listeners.shutdown()

    #     from bitdust.main import events
    #     events.shutdown()

    #     from bitdust.automats import automat
    #     automat.objects().clear()
    #     if len(automat.index()) > 0:
    #         lg.warn('%d automats was not cleaned' % len(automat.index()))
    #         for a in automat.index().keys():
    #             if _Debug:
    #                 lg.out(_DebugLevel, '    %r' % a)
    #     else:
    #         if _Debug:
    #             lg.out(_DebugLevel, 'bpmain.shutdown automat.objects().clear() SUCCESS, no state machines left in memory')

    if _Debug:
        lg.out(_DebugLevel, 'bpmain.shutdown currently %d threads running:' % len(threading.enumerate()))
    for t in threading.enumerate():
        if _Debug:
            lg.out(_DebugLevel, '    ' + str(t))

    if _Debug:
        lg.out(_DebugLevel, 'bpmain.shutdown finishing and closing log file, EXIT')

    shutdowner.shutdown_settings()

    # automat.CloseLogFile()
    # automat.SetExceptionsHandler(None)
    # automat.SetLogOutputHandler(None)

    lg.close_log_file()
    lg.close_intercepted_log_file()

    lg.stdout_stop_redirecting()
    lg.stderr_stop_redirecting()

    # from bitdust.main import settings
    # settings.shutdown()

    return 0


#------------------------------------------------------------------------------


def run_twisted_reactor():
    from bitdust.logs import lg
    try:
        from twisted.internet import reactor  # @UnresolvedImport
    except:
        lg.exc()
        sys.exit('Error initializing reactor in bpmain.py\n')
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.run_twisted_reactor calling Twisted reactor.run()')
    reactor.run()  # @UndefinedVariable
    if _Debug:
        lg.out(_DebugLevel, 'bpmain.run_twisted_reactor Twisted reactor stopped')


def run(UI='', options=None, args=None, overDict=None, executablePath=None, start_reactor=True):
    if options and options.cpu_profile:
        import cProfile, pstats, io
        from pstats import SortKey  # @UnresolvedImport
        pr = cProfile.Profile()
        pr.enable()

    init(UI, options, args, overDict, executablePath)

    if start_reactor:
        run_twisted_reactor()
        result = shutdown()
    else:
        result = True

    if options and options.cpu_profile:
        pr.disable()
        s = io.StringIO()
        sortby = SortKey.CUMULATIVE
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        open('/tmp/bitdust.profile', 'w').write(s.getvalue())

    return result


#------------------------------------------------------------------------------


def parser():
    """
    Create an ``optparse.OptionParser`` object to read command line arguments.
    """
    from optparse import OptionParser, OptionGroup
    parser = OptionParser(usage=usage_text(), prog='BitDust')
    group = OptionGroup(parser, 'Logs')
    group.add_option(
        '-d',
        '--debug',
        dest='debug',
        type='int',
        help='set debug level',
    )
    group.add_option(
        '-q',
        '--quite',
        dest='quite',
        action='store_true',
        help='quite mode, do not print any messages to stdout',
    )
    group.add_option(
        '-v',
        '--verbose',
        dest='verbose',
        action='store_true',
        help='verbose mode, print more messages',
    )
    group.add_option(
        '--coverage',
        dest='coverage',
        action='store_true',
        help='record code coverage',
    )
    group.add_option('--coverage_config', dest='coverage_config', type='string', help='coverage configuration file path')
    group.add_option('--coverage_report', dest='coverage_report', type='string', help='file path to be used to store coverage report')
    group.add_option(
        '-n',
        '--no-logs',
        dest='no_logs',
        action='store_true',
        help='do not use logs',
    )
    group.add_option(
        '-o',
        '--output',
        dest='output',
        type='string',
        help='print log messages to the file',
    )
    group.add_option(
        '-a',
        '--appdir',
        dest='appdir',
        type='string',
        help='set alternative location for application data files, default is ~/.bitdust/',
    )
    #    group.add_option('-t', '--tempdir',
    #                        dest='tempdir',
    #                        type='string',
    #                        help='set location for temporary files, default is ~/.bitdust/temp',)
    group.add_option(
        '--twisted',
        dest='twisted',
        action='store_true',
        help='show twisted log messages too',
    )
    group.add_option(
        '--cpu-profile',
        dest='cpu_profile',
        action='store_true',
        help='use cProfile to profile performance, output is in the file /tmp/bitdust.profile',
    )
    #    group.add_option('--memdebug',
    #                        dest='memdebug',
    #                        action='store_true',
    #                        help='start web server to debug memory usage, need cherrypy and dozer modules',)
    parser.add_option_group(group)
    #    group = OptionGroup(parser, "Network")
    #    group.add_option('--tcp-port',
    #                        dest='tcp_port',
    #                        type='int',
    #                        help='set tcp port number for incoming connections',)
    #    group.add_option('--no-upnp',
    #                        dest='no_upnp',
    #                        action='store_true',
    #                        help='do not use UPnP',)
    #    group.add_option('--memdebug-port',
    #                        dest='memdebug_port',
    #                        type='int',
    #                        default=9996,
    #                        help='set port number for memdebug web server, default is 9995',)
    #    parser.add_option_group(group)
    return parser


def override_options(opts, args):
    """
    The program can replace some user options by values passed via command
    line.

    This method return a dictionary where is stored a key-value pairs
    for new options.
    """
    overDict = {}
    #    if opts.tcp_port:
    #        overDict['services/tcp-connections/tcp-port'] = str(opts.tcp_port)
    #    if opts.no_upnp:
    #        overDict['services/tcp-connections/upnp-enabled'] = 'false'
    if opts.debug or str(opts.debug) == '0':
        overDict['logs/debug-level'] = str(opts.debug)
#    if opts.memdebug:
#        overDict['logs/memdebug-enable'] = str(opts.memdebug)
#        if opts.memdebug_port:
#            overDict['logs/memdebug-port'] = str(opts.memdebug_port)
#        else:
#            overDict['logs/memdebug-port'] = '9996'
    return overDict


#------------------------------------------------------------------------------


def kill():
    """
    Kill all running BitDust processes (except current).
    """
    from bitdust.system import bpio
    total_count = 0
    found = False
    while True:
        appList = bpio.lookup_main_process()
        if len(appList) > 0:
            found = True
        for pid in appList:
            print_text('trying to kill process %d' % pid)
            bpio.kill_process(pid)
        if len(appList) == 0:
            if found:
                print_text('BitDust stopped\n', nl='')
            else:
                print_text('BitDust was not started\n', nl='')
            return 0
        total_count += 1
        if total_count > 10:
            print_text('some BitDust process found, but can not be stoped\n', nl='')
            return 1
        time.sleep(1)


def wait_then_kill(x):
    """
    For correct shutdown of the program need to send a URL request to the HTTP
    server:: http://localhost:<random port>/?action=exit.

    After receiving such request the program will call
    ``p2p.init_shutdown.shutdown()`` method and stops. But if the main
    process was blocked it needs to be killed with system "kill"
    procedure. This method will wait for 10 seconds and then call method
    ``kill()``.
    """
    from twisted.internet import reactor  # @UnresolvedImport
    from bitdust.system import bpio
    total_count = 0
    while True:
        appList = bpio.lookup_main_process()
        if len(appList) == 0:
            print_text('DONE')
            reactor.stop()  # @UndefinedVariable
            return 0
        total_count += 1
        if total_count > 10:
            print_text('not responding, killing the process now ...')
            ret = kill()
            reactor.stop()  # @UndefinedVariable
            return ret
        time.sleep(1)


#------------------------------------------------------------------------------

_OriginalCallLater = None
_LastCallableID = 0


class _callable():
    """
    This class shows my experiments with performance monitoring.

    I tried to decrease the number of delayed calls.
    """
    def __init__(self, delay, callabl, *args, **kw):
        self.callabl = callabl
        self.to_call = lambda: self.run(*args, **kw)

    def run(self, *args, **kw):
        from bitdust.logs import measure_it
        measure_it.run(self.callabl, *args, **kw)

    def call(self):
        self.to_call()


def _callLater(delay, callabl, *args, **kw):
    """
    A wrapper around Twisted ``reactor.callLater()`` method.
    """
    global _OriginalCallLater
    _call = _callable(delay, callabl, *args, **kw)
    delayed_call = _OriginalCallLater(delay, _call.call)
    return delayed_call


def patchReactorCallLater(r, apply=True):
    """
    Replace original ``reactor.callLater()`` with my hacked solution to monitor
    overall performance.
    """
    global _OriginalCallLater
    if apply:
        _OriginalCallLater = r.callLater
        r.callLater = _callLater
    else:
        r.callLater = _OriginalCallLater
        _OriginalCallLater = None


def monitorDelayedCalls(r):
    """
    Print out all delayed calls.
    """
    from bitdust.logs import measure_it
    from bitdust.logs import lg
    stats = measure_it.top_calls()
    if _Debug:
        lg.out(_DebugLevel, '\nslowest calls:\n%s' % stats)
    r.callLater(30, monitorDelayedCalls, r)


#------------------------------------------------------------------------------


class TwistedUnhandledErrorsObserver:
    def __init__(self, level):
        self.level = level

    def __call__(self, event_dict):
        if event_dict.get('log_level') == self.level:
            if 'log_failure' in event_dict:
                f = event_dict['log_failure']
                from bitdust.logs import lg
                lg.exc(msg=f'Unhandled error in Deferred:\n{event_dict.get("debugInfo", "")}', exc_info=(f.type, f.value, f.getTracebackObject()))


#-------------------------------------------------------------------------------


def usage_text():
    """
    Calls ``p2p.help.usage_text()`` method to print out how to run BitDust software
    from command line.
    """
    try:
        from bitdust.main import help
        return help.usage_text()
    except:
        return ''


def help_text():
    """
    Same thing, calls ``p2p.help.help_text()`` to show detailed instructions.
    """
    try:
        from bitdust.main import help
        return help.help_text()
    except:
        return ''


def backup_schedule_format():
    """
    See ``p2p.help.schedule_format()`` method.
    """
    try:
        from bitdust.main import help
        return help.schedule_format()
    except:
        return ''


def copyright_text():
    """
    Prints the copyright string.
    """
    print('Copyright (C) 2008 Veselin Penev, https://bitdust.io')


#--- THE ENTRY POINT
def main(executable_path=None, start_reactor=True, appdir=None):
    """
    THE ENTRY POINT
    """
    global _AppDataDir

    if _Debug:
        print_text('ENTRY POINT: executable_path=%s appdir=%s' % (
            executable_path,
            appdir,
        ))

    pars = parser()
    (opts, args) = pars.parse_args()

    if not appdir:
        appdir = opts.appdir

    if opts.coverage:
        import coverage  # @UnresolvedImport
        cov = coverage.Coverage(config_file=opts.coverage_config)
        cov.start()

    overDict = override_options(opts, args)

    cmd = ''
    if len(args) > 0:
        cmd = args[0].lower()

    try:
        from bitdust.system import deploy
    except:
        dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
        sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
        from distutils.sysconfig import get_python_lib
        sys.path.append(os.path.join(get_python_lib(), 'bitdust'))
        try:
            from bitdust.system import deploy
        except:
            print_text('ERROR! can not import working code.  Python Path:')
            print_text('\n'.join(sys.path))
            return 1

    #---install---
    if cmd in ['deploy', 'install', 'venv', 'virtualenv']:
        from bitdust.system import deploy
        return deploy.run(args)

    #---integrate---
    if cmd == 'integrate' or cmd == 'alias' or cmd == 'shell':
        """
        This is a helper command to make a "system-wide" command called "bitdust" for fast access to the app.

        Run:
            python3 bitdust.py alias > /usr/local/bin/bitdust
            chmod +x /usr/local/bin/bitdust

        This will create an executable file /usr/local/bin/bitdust with such content:
            #!/bin/sh
            python3 [path to `bitdust` folder]/bitdust.py "$@"
        """
        from bitdust.system import bpio
        if bpio.Windows():
            # TODO:
            # src = u"""@echo off
            # C:\Users\veselin\BITDUS~2\venv\Scripts\python.exe C:\Users\veselin\BITDUS~2\src\bitdust.py %*
            # """
            print_text('this feature is not yet available for Windows')
            return 0
        python_path = os.path.abspath(os.path.join(os.path.expanduser('~'), '.bitdust', 'venv', 'bin', 'python'))
        curpath = os.path.abspath(os.path.join(bpio.getExecutableDir(), '..'))
        src = '#!/bin/sh\n'
        src += '# This is a very short shell script to help you create an alias "bitdust" in your OS for the BitDust software.\n'
        src += '# NOTICE: BitDust do not require root permissions to run, please always start it as normal user.\n'
        src += '# Run:\n'
        src += '#     python3 bitdust.py alias > /usr/local/bin/bitdust\n'
        src += '#     chmod +x /usr/local/bin/bitdust\n'
        src += '%s %s/bitdust.py "$@"\n' % (python_path, curpath)
        print_text(src)
        return 0

    if appdir:
        appdata = appdir
        _AppDataDir = appdata

    else:
        curdir = os.getcwd()
        appdatafile = os.path.join(curdir, 'appdata')
        defaultappdata = deploy.default_base_dir_portable()
        appdata = defaultappdata
        if os.path.isfile(appdatafile):
            try:
                appdata = os.path.abspath(open(appdatafile, 'rb').read().strip())
            except:
                appdata = defaultappdata
            if not os.path.isdir(appdata):
                appdata = defaultappdata
        _AppDataDir = appdata

    if _Debug:
        print_text('_AppDataDir: %s' % _AppDataDir)
        print_text('default_base_dir_portable(): %s' % deploy.default_base_dir_portable())

    #---BitDust Home
    try:
        result_app_dir = deploy.init_base_dir(base_dir=_AppDataDir)
    except Exception as e:
        print_text('failed to initialize application data folder: %e' % e)
        return 1

    if _Debug:
        print_text('app data dir prepared in %s' % result_app_dir)

    from bitdust.logs import lg

    #---init IO module
    from bitdust.system import bpio
    bpio.init()

    appList = bpio.find_main_process(pid_file_path=os.path.join(appdata, 'processid'))

    if bpio.Android():
        from android.storage import app_storage_path  # @UnresolvedImport
        android_log_file = os.path.join(app_storage_path(), '.bitdust', 'logs', 'android.log')
        if _Debug:
            print_text('redirecting log output to %s' % android_log_file)
        lg.close_intercepted_log_file()
        lg.open_intercepted_log_file(android_log_file)

    # sys.excepthook = lg.exception_hook

    #---init logging
    from twisted.internet.defer import setDebugging
    if _Debug:
        if bpio.isFrozen():
            setDebugging(False)
        else:
            setDebugging(True)
    else:
        setDebugging(False)

    from twisted.logger import globalLogPublisher, LogLevel
    tw_log_observer = TwistedUnhandledErrorsObserver(level=LogLevel.critical)
    globalLogPublisher.addObserver(tw_log_observer)

    #---life begins!
    # ask logger to count time for each log line from that moment, not absolute time
    lg.life_begins()

    # try to read debug level value at the early stage - no problem if fail here
    try:
        if cmd == '' or cmd == 'start' or cmd == 'go' or cmd == 'show' or cmd == 'open':
            lg.set_debug_level(int(bpio.ReadTextFile(os.path.abspath(os.path.join(appdata, 'config', 'logs', 'debug-level')))))
    except:
        pass

    if opts.no_logs:
        lg.disable_logs()

    if opts.debug or str(opts.debug) == '0':
        lg.set_debug_level(int(opts.debug))

    #---logpath---
    logpath = None
    if opts.output:
        logpath = opts.output
    else:
        try:
            os.makedirs(os.path.join(appdata, 'logs'), exist_ok=True)
        except:
            pass
        logpath = os.path.join(appdata, 'logs', 'stdout.log')

    need_redirecting = False

    if bpio.Windows() and not bpio.isConsoled():
        need_redirecting = True

    if logpath:
        if not appList:
            if cmd not in [
                'detach',
                'daemon',
                'stop',
                'kill',
                'shutdown',
                'restart',
                'reboot',
                'reconnect',
                'show',
                'open',
            ]:
                lg.open_log_file(logpath)
        if bpio.Windows() and bpio.isFrozen():
            need_redirecting = True

    if bpio.Android():
        need_redirecting = True

    if opts.quite and not opts.verbose:
        lg.disable_output()
    else:
        if need_redirecting:
            lg.stdout_start_redirecting()
            lg.stderr_start_redirecting()

    #---start---
    if cmd == '' or cmd == 'start' or cmd == 'go':
        if appList:
            print_text('BitDust already started, found another process: %s\n' % str(appList), nl='')
            bpio.shutdown()
            return 0

        UI = ''
        try:
            ret = run(UI, opts, args, overDict, executable_path, start_reactor)
        except:
            lg.exc()
            ret = 1
        bpio.shutdown()

        if opts.coverage:
            cov.stop()
            cov.save()
            if opts.coverage_report:
                cov.report(file=open(opts.coverage_report, 'w'))

        return ret

    #---daemon---
    elif cmd == 'detach' or cmd == 'daemon':
        appList = bpio.find_main_process(pid_file_path=os.path.join(appdata, 'processid'))
        if len(appList) > 0:
            print_text('main BitDust process already started: %s\n' % str(appList), nl='')
            bpio.shutdown()
            if opts.coverage:
                cov.stop()
                cov.save()
                if opts.coverage_report:
                    cov.report(file=open(opts.coverage_report, 'w'))
            return 0
        from bitdust.lib import misc
        print_text('new BitDust process will be started in daemon mode\n', nl='')
        result = misc.DoRestart(detach=True,
                                # std_out=os.path.join(appdata, 'logs', 'stdout.log'),
                                # std_err=os.path.join(appdata, 'logs', 'stderr.log'),
                               )
        if result is not None:
            try:
                result = int(result)
            except:
                try:
                    result = result.pid
                except:
                    pass
        bpio.shutdown()
        if opts.coverage:
            cov.stop()
            cov.save()
            if opts.coverage_report:
                cov.report(file=open(opts.coverage_report, 'w'))
        return 0

    #---restart---
    elif cmd == 'restart' or cmd == 'reboot':
        appList = bpio.find_main_process(pid_file_path=os.path.join(appdata, 'processid'))
        ui = False
        if len(appList) > 0:
            print_text('found main BitDust process: %r ... ' % appList, nl='')

            def done(x):
                print_text('finished successfully\n', nl='')
                from twisted.internet import reactor  # @UnresolvedImport
                if reactor.running and not reactor._stopped:  # @UndefinedVariable
                    reactor.stop()  # @UndefinedVariable

            def failed(x):
                if isinstance(x, Failure):
                    print_text('finished with: %s\n' % x.getErrorMessage(), nl='')
                else:
                    print_text('finished successfully\n', nl='')
                ok = str(x).count('Connection was closed cleanly') > 0
                from twisted.internet import reactor  # @UnresolvedImport
                if ok and reactor.running and not reactor._stopped:  # @UndefinedVariable
                    # print_text('DONE\n', '')
                    reactor.stop()  # @UndefinedVariable
                    return
                print_text('forcing previous process shutdown\n', nl='')
                try:
                    kill()
                except:
                    lg.exc()
                from bitdust.lib import misc
                reactor.addSystemEventTrigger(  # @UndefinedVariable
                    'after',
                    'shutdown',
                    misc.DoRestart,
                    param='show' if ui else '',
                    detach=True,  # std_out=os.path.join(appdata, 'logs', 'stdout.log'),
                    # std_err=os.path.join(appdata, 'logs', 'stderr.log'),
                )
                reactor.stop()  # @UndefinedVariable

            try:
                from twisted.internet import reactor  # @UnresolvedImport
                # from interface.command_line import run_url_command
                # d = run_url_command('?action=restart', False)
                # from bitdust.interface import cmd_line
                # d = cmd_line.call_xmlrpc_method('restart', ui)
                from bitdust.interface import cmd_line_json
                d = cmd_line_json.call_websocket_method('process_restart', websocket_timeout=5)
                d.addCallback(done)
                d.addErrback(failed)
                reactor.run()  # @UndefinedVariable
                bpio.shutdown()
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return 0
            except:
                lg.exc()
                bpio.shutdown()
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return 1
        else:
            ui = ''
            if cmd == 'restart':
                ui = 'show'
            try:
                ret = run(ui, opts, args, overDict, executable_path)
            except:
                lg.exc()
                ret = 1
            bpio.shutdown()
            if opts.coverage:
                cov.stop()
                cov.save()
                if opts.coverage_report:
                    cov.report(file=open(opts.coverage_report, 'w'))
            return ret

    #---show---
    elif cmd == 'show' or cmd == 'open':
        if not bpio.isGUIpossible():
            print_text('BitDust GUI is turned OFF\n', nl='')
            bpio.shutdown()
            return 0
        if bpio.Linux() and not bpio.X11_is_running():
            print_text('this operating system not supporting X11 interface\n', nl='')
            bpio.shutdown()
            return 0
        appList = bpio.find_main_process(pid_file_path=os.path.join(appdata, 'processid'))
        if len(appList) == 0:
            try:
                ret = run('show', opts, args, overDict, executable_path)
            except:
                lg.exc()
                ret = 1
            bpio.shutdown()
            return ret
        # print_text('found main BitDust process: %s, start the GUI\n' % str(appList))
        # ret = show()
        bpio.shutdown()
        return ret

    #---stop---
    elif cmd == 'stop' or cmd == 'kill' or cmd == 'shutdown':
        if cmd == 'kill':
            ret = kill()
            bpio.shutdown()
            if opts.coverage:
                cov.stop()
                cov.save()
                if opts.coverage_report:
                    cov.report(file=open(opts.coverage_report, 'w'))
            return ret
        appList = bpio.find_main_process(pid_file_path=os.path.join(appdata, 'processid'))
        if len(appList) > 0:
            if cmd == 'kill':
                print_text('found main BitDust process: %s, about to kill running process ... ' % appList, nl='')
                ret = kill()
                bpio.shutdown()
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return ret
            try:
                from twisted.internet import reactor  # @UnresolvedImport
                from twisted.python.failure import Failure

                def _stopped(x):
                    if _Debug:
                        if isinstance(x, Failure):
                            print_text('finished with: %s\n' % x.getErrorMessage(), nl='')
                        else:
                            print_text('finished with: %s\n' % x, nl='')
                    else:
                        print_text('finished successfully\n', nl='')
                    reactor.stop()  # @UndefinedVariable
                    bpio.shutdown()

                print_text('found main BitDust process: %s ... ' % appList, nl='')
                from bitdust.interface import cmd_line_json
                cmd_line_json.call_websocket_method('process_stop', websocket_timeout=2).addBoth(_stopped)
                reactor.run()  # @UndefinedVariable
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return 0
            except:
                lg.exc()
                ret = kill()
                bpio.shutdown()
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return ret
        else:
            appListAllChilds = bpio.find_main_process(
                check_processid_file=False,
                extra_lookups=[],
            )
            if len(appListAllChilds) > 0:
                print_text('BitDust child processes found: %s, performing "kill process" action ...\n' % appListAllChilds, nl='')
                ret = kill()
                if opts.coverage:
                    cov.stop()
                    cov.save()
                    if opts.coverage_report:
                        cov.report(file=open(opts.coverage_report, 'w'))
                return ret

            print_text('BitDust is not running at the moment\n', nl='')
            bpio.shutdown()
            if opts.coverage:
                cov.stop()
                cov.save()
                if opts.coverage_report:
                    cov.report(file=open(opts.coverage_report, 'w'))
            return 0

    #---command_line---
    from bitdust.interface import cmd_line_json as cmdln
    ret = cmdln.run(opts, args, pars, overDict, executable_path)
    if ret == 2:
        print_text(usage_text())
    bpio.shutdown()

    #---coverage report---
    if opts.coverage:
        cov.stop()
        cov.save()
        if opts.coverage_report:
            cov.report(file=open(opts.coverage_report, 'w'))

    return ret


#------------------------------------------------------------------------------

if __name__ == '__main__':
    ret = main()
    if ret == 2:
        print(usage_text())



#!/usr/bin/python
# ftp_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (ftp_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
from __future__ import absolute_import
"""
..

module:: ftp_server
"""

#------------------------------------------------------------------------------

import os
import time

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet import defer
from twisted.internet.defer import Deferred, succeed

from twisted.cred.portal import Portal
from twisted.cred.checkers import AllowAnonymousAccess, FilePasswordDB

from twisted.python import filepath

from twisted.protocols.ftp import (
    # classes
    FTPFactory,
    FTPRealm,
    FTP,
    ASCIIConsumerWrapper,
    _FileReader,
    _FileWriter,  # methods
    toSegments,
    errnoToFailure,  # exceptions
    FTPCmdError,
    BadCmdSequenceError,
    InvalidPath,
    FileNotFoundError,
    FileExistsError,
    PermissionDeniedError,
    IsADirectoryError,  # FTP codes constants
    DATA_CNX_ALREADY_OPEN_START_XFR,
    TXFR_COMPLETE_OK,
    REQ_FILE_ACTN_COMPLETED_OK,
    CNX_CLOSED_TXFR_ABORTED,
    FILE_NOT_FOUND,
    FILE_STATUS_OK_OPEN_DATA_CNX,
    FILE_STATUS,
    MKD_REPLY,
)

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings

from bitdust.system import tmpfile
from bitdust.system import bpio

from bitdust.storage import backup_fs
from bitdust.storage import backup_control

from bitdust.interface import api

#------------------------------------------------------------------------------

_FTPServer = None

#------------------------------------------------------------------------------


def init(ftp_port=None):
    global _FTPServer
    lg.out(4, 'ftp_server.init')
    if _FTPServer:
        lg.warn('already started')
        return
    if not ftp_port:
        ftp_port = settings.getFTPServerPort()
    if not os.path.isfile(settings.FTPServerCredentialsFile()):
        bpio.WriteTextFile(settings.FTPServerCredentialsFile(), 'bitdust:bitdust')
    # TODO: add protection: accept connections only from local host: 127.0.0.1
    _FTPServer = reactor.listenTCP(  # @UndefinedVariable
        ftp_port, BitDustFTPFactory(Portal(
            FTPRealm('./'),
            [
                AllowAnonymousAccess(),
                FilePasswordDB(settings.FTPServerCredentialsFile()),
            ],
        ))
    )
    lg.out(4, '    started on port %d' % ftp_port)


def shutdown():
    global _FTPServer
    lg.out(4, 'ftp_server.shutdown')
    if not _FTPServer:
        return succeed(None)
    result = Deferred()
    lg.out(4, '    calling stopListening()')
    _FTPServer.stopListening().addBoth(lambda *args: result.callback(*args))
    _FTPServer = None
    return result


#------------------------------------------------------------------------------


class BitDustFileReader(_FileReader):
    def __init__(self, fObj, filepath):
        super(BitDustFileReader, self).__init__(fObj)
        self.filepath = filepath

    def _close(self, passthrough):
        passthrough = super(BitDustFileReader, self)._close(passthrough)
        # tmpfile.throw_out(os.path.dirname(self.filepath), 'file read')
        return passthrough


#------------------------------------------------------------------------------


class BitDustFTP(FTP):
    def _accessGrantedResponse(self, result, segments):
        self.workingDirectory = segments
        return (REQ_FILE_ACTN_COMPLETED_OK, )

    def _dirListingResponse(self, results):
        self.reply(DATA_CNX_ALREADY_OPEN_START_XFR)
        for (name, attrs) in results:
            name = self._encodeName(name)
            self.dtpInstance.sendListResponse(name, attrs)
        self.dtpInstance.transport.loseConnection()
        return (TXFR_COMPLETE_OK, )

    def _enableTimeoutLater(self, result, newsegs=None):
        self.setTimeout(self.factory.timeOut)
        return result

    def _cbFileSent(self, result):
        lg.out(8, 'ftp_server._cbFileSent %s' % result)
        return (TXFR_COMPLETE_OK, )

    def _ebFileSent(self, err):
        lg.warn('Unexpected error attempting to transmit file to client: ' + str(err))
        if err.check(FTPCmdError):
            return err
        return (CNX_CLOSED_TXFR_ABORTED, )

    def _cbReadOpened(self, file_obj, consumer):
        lg.out(8, 'ftp_server._cbWriteOpened %s %s' % (file_obj, consumer))
        if self.dtpInstance.isConnected:
            self.reply(DATA_CNX_ALREADY_OPEN_START_XFR)
        else:
            self.reply(FILE_STATUS_OK_OPEN_DATA_CNX)
        d = file_obj.send(consumer)
        d.addCallbacks(self._cbFileSent, self._ebFileSent)
        return d

    def _ebReadOpened(self, err, newsegs):
        if not err:
            return (FILE_NOT_FOUND, '/'.join(newsegs))
        if not err.check(PermissionDeniedError, FileNotFoundError, IsADirectoryError):
            lg.warn('Unexpected error attempting to open file for transmission: ' + str(err))
        if err.check(FTPCmdError):
            return (err.value.errorCode, '/'.join(newsegs))
        return (FILE_NOT_FOUND, '/'.join(newsegs))

    def _cbFileRecevied(self, consumer, local_path, newsegs):
        #         receive_defer.addCallback(self._startFileBackup, upload_filename, newsegs, d)
        #         consumer.fObj.flush()
        #         os.fsync(consumer.fObj.fileno())
        #         consumer.fObj.close()
        #         consumer.close()
        remote_path = '/'.join(newsegs)
        lg.out(8, 'ftp_server._cbFileRecevied %s %s' % (local_path, remote_path))
        ret = api.file_info(remote_path)
        if ret['status'] != 'OK':
            ret = api.file_create(remote_path)
            if ret['status'] != 'OK':
                return defer.fail(FileNotFoundError(remote_path))
        else:
            if ret['result']['type'] == 'dir':
                return defer.fail(IsADirectoryError(remote_path))
        ret = api.file_upload_start(local_path, remote_path, wait_result=False)
        if ret['status'] != 'OK':
            lg.warn('file_upload_start() returned: %s' % ret)
            return defer.fail(FileNotFoundError(remote_path))

#         shortPathID = backup_fs.ToID(full_path)
#         if not shortPathID:
#             shortPathID, _, _ = backup_fs.AddFile(full_path, read_stats=False)
#         item = backup_fs.GetByID(shortPathID)
#         item.read_stats(upload_filename)
#         backup_control.StartSingle(shortPathID, upload_filename)
#         # upload_task.result_defer.addCallback(self._cbFileBackup, result_defer, newsegs)
#         backup_fs.Calculate()
#         backup_control.Save()
# result_defer.callback(None)
# return consumer
        return (TXFR_COMPLETE_OK, )

    def _ebFileReceived(self, err):
        lg.warn('Unexpected error received during transfer: ' + str(err))
        if err.check(FTPCmdError):
            return err
        return (CNX_CLOSED_TXFR_ABORTED, )

    def _cbWriteOpened(self, consumer, upload_filename, newsegs):
        remote_path = '/'.join(newsegs)
        lg.out(8, 'ftp_server._cbWriteOpened %s %s' % (upload_filename, remote_path))
        d = consumer.receive()
        d.addCallback(self._startConsumer)
        d.addCallback(self._stopConsumer, consumer)
        # d.addCallback(lambda ignored: file_writer.close())
        d.addCallback(self._cbFileRecevied, upload_filename, newsegs)
        d.addErrback(self._ebFileReceived)
        return d

    def _ebWriteOpened(self, err, newsegs):
        if isinstance(err.value, FTPCmdError):
            return (err.value.errorCode, '/'.join(newsegs))
        lg.warn('Unexpected error received while opening file: %s' % str(err))
        return (FILE_NOT_FOUND, '/'.join(newsegs))

    def _startConsumer(self, consumer):
        if not self.binary:
            consumer = ASCIIConsumerWrapper(consumer)
        d = self.dtpInstance.registerConsumer(consumer)
        if self.dtpInstance.isConnected:
            self.reply(DATA_CNX_ALREADY_OPEN_START_XFR)
        else:
            self.reply(FILE_STATUS_OK_OPEN_DATA_CNX)
        return d

    def _stopConsumer(self, d, consumer):
        # consumer.fObj.flush()
        # os.fsync(consumer.fObj.fileno())
        # consumer.fObj.close()
        consumer.close()
        return d

    def _cbRestoreDone(self, ret, path_segments, result_defer):
        pth = '/'.join(path_segments)
        lg.out(8, 'ftp_server._cbRestoreDone %s %s' % (ret, pth))
        if ret['status'] != 'OK':
            return result_defer.errback(FileNotFoundError(pth))
        if not ret['result']['downloaded']:
            return result_defer.errback(FileNotFoundError(pth))
        fp = filepath.FilePath(os.path.join(ret['local_path'], os.path.basename(ret['remote_path'])))
        try:
            fobj = fp.open('r')
        except:
            return result_defer.errback(FileNotFoundError(pth))
        fr = _FileReader(fobj)
        return result_defer.callback(fr)

#     def _cbStat(self, result):
#         (size,) = result
#         return (FILE_STATUS, str(size))

#------------------------------------------------------------------------------

    def ftp_LIST(self, path=''):
        # Uh, for now, do this retarded thing.
        if self.dtpInstance is None or not self.dtpInstance.isConnected:
            return defer.fail(BadCmdSequenceError('must send PORT or PASV before RETR'))
        # Various clients send flags like -L or -al etc.  We just ignore them.
        if path.lower() in ['-a', '-l', '-la', '-al']:
            path = ''
        try:
            segments = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        pth = '/'.join(segments)
        ret = api.files_list(pth)
        if ret['status'] != 'OK':
            return defer.fail(FileNotFoundError(path))
        lst = ret['result']
        result = []
        for itm in lst:
            if itm['path'] == 'index':
                continue
            # known_size = max(itm[7].size, 0)
#             if itm['versions']:
#                 known_size = itm['size']
#             else:
#                 known_size = 1
            known_size = max(itm['local_size'], 0)
            key_alias, _, _ = itm['key_id'].partition('$')
            result.append((
                os.path.basename(itm['path']),
                [  # name
                    known_size,  # size
                    True if itm['type'] == 'dir' else False,  # folder or file ?
                    filepath.Permissions(0o7777),  # permissions
                    0,  # hardlinks
                    time.mktime(time.strptime(itm['latest'], '%Y-%m-%d %H:%M:%S'), ) if itm['latest'] else None,  # time
                    itm['customer'],  # owner
                    key_alias,  # group
                ],
            ))
        d = Deferred()
        d.addCallback(self._dirListingResponse)
        d.callback(result)
        return d

    def ftp_CWD(self, path):
        try:
            segments = toSegments(self.workingDirectory, path)
        except InvalidPath:
            # XXX Eh, what to fail with here?
            return defer.fail(FileNotFoundError(path))
        pth = '/'.join(segments)
        d = Deferred()
        d.addCallback(lambda r: self._accessGrantedResponse(r, segments))
        if not pth or pth == '/':
            d.callback(None)
            return d
        ret = api.file_info(pth, include_uploads=False, include_downloads=False)
        if ret['status'] != 'OK':
            d.errback(FileNotFoundError(path))
            return d
        if ret['result']['type'] == 'dir':
            d.callback(None)
        else:
            d.errback(FileNotFoundError(path))
        return d

    def ftp_RETR(self, path):
        if self.dtpInstance is None:
            raise BadCmdSequenceError('PORT or PASV required before RETR')
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        # XXX For now, just disable the timeout.  Later we'll want to
        # leave it active and have the DTP connection reset it
        # periodically.
        self.setTimeout(None)
        # Put it back later
        # And away she goes
        if not self.binary:
            consumer = ASCIIConsumerWrapper(self.dtpInstance)
        else:
            consumer = self.dtpInstance
        pth = '/'.join(newsegs)
        restore_dir = tmpfile.make_dir('restore', prefix=('_'.join(newsegs) + '_'))
        ret = api.file_download_start(pth, restore_dir, wait_result=True)
        d = Deferred()
        d.addCallback(self._cbReadOpened, consumer)
        d.addErrback(self._ebReadOpened, newsegs)
        d.addBoth(self._enableTimeoutLater)
        if isinstance(ret, dict):
            #             if ret['status'] != 'OK':
            #                 return defer.fail(FileNotFoundError(path))
            self._cbRestoreDone(ret, newsegs, d)
            return d
        ret.addCallback(self._cbRestoreDone, newsegs, d)
        ret.addErrback(lambda err: lg.exc(err))
        return d

    def ftp_STOR(self, path):
        if self.dtpInstance is None:
            raise BadCmdSequenceError('PORT or PASV required before STOR')
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))

#         parent_path = '/' + ('/'.join(newsegs[:-1]))
#         parent_item = backup_fs.GetByPath(parent_path)
#         if not parent_item:
#             return defer.fail(FileNotFoundError(parent_path))

# XXX For now, just disable the timeout.  Later we'll want to
# leave it active and have the DTP connection reset it
# periodically.
        self.setTimeout(None)
        # Put it back later

        # , prefix=(newsegs[-1] + '_')
        upload_dir = tmpfile.make_dir('backup', extension='.ftp')
        if not upload_dir:
            return defer.fail(FileNotFoundError(path))
        upload_filename = os.path.join(upload_dir, newsegs[-1])
        fp = filepath.FilePath(upload_filename)
        try:
            fObj = fp.open('w')
        except (IOError, OSError) as e:
            return errnoToFailure(e.errno, path)
        except:
            return defer.fail(FileNotFoundError(path))
        # TODO: remove file after all
        # d.addCallback(lambda ignored: file.close(); and remove file)
        # d = Deferred()
        d = Deferred()
        d.addCallback(self._cbWriteOpened, upload_filename, newsegs)
        d.addErrback(self._ebWriteOpened, newsegs)
        d.addBoth(self._enableTimeoutLater)
        d.callback(_FileWriter(fObj))

        #         d.addCallbacks(self._cbFileRecevied, self._ebFileReceived)
        #         fw = _FileWriter(fObj)
        #         receive_defer = fw.receive()
        #         receive_defer.addBoth(self._enableTimeoutLater)
        #         receive_defer.addCallback(self._prepareConsumer)
        #         receive_defer.addCallback(self._startFileBackup, upload_filename, newsegs, d)
        #         receive_defer.addErrback(lambda err: d.errback(FileNotFoundError(path)))

        return d

    def ftp_SIZE(self, path):
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        full_path = '/'.join(newsegs)
        ret = api.file_info(full_path)
        if ret['status'] != 'OK':
            return defer.fail(FileNotFoundError(path))
        return succeed((
            FILE_STATUS,
            str(ret['size']),
        ))

#         shortPathID = backup_fs.ToID(full_path)
#         if shortPathID is None:
#             return defer.fail(FileNotFoundError(path))
#         item = backup_fs.GetByID(shortPathID)
#         if item is None:
#             return defer.fail(FileNotFoundError(path))
#         return succeed((FILE_STATUS, str(item.size), ))

    def ftp_MKD(self, path):
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        full_path = '/'.join(newsegs)
        ret = api.file_create(full_path, as_folder=True)
        if ret['status'] != 'OK':
            return defer.fail(FileExistsError(str(ret['errors'])))
        return succeed((MKD_REPLY, path))

    def ftp_RMD(self, path):
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        full_path = '/'.join(newsegs)
        ret = api.file_delete(full_path)
        if ret['status'] != 'OK':
            return defer.fail(FileNotFoundError(str(ret['errors'])))
        return succeed((REQ_FILE_ACTN_COMPLETED_OK, ))

    def ftp_DELE(self, path):
        try:
            newsegs = toSegments(self.workingDirectory, path)
        except InvalidPath:
            return defer.fail(FileNotFoundError(path))
        full_path = '/'.join(newsegs)
        ret = api.file_delete(full_path)
        if ret['status'] != 'OK':
            return defer.fail(FileNotFoundError(str(ret['errors'])))
        return succeed((REQ_FILE_ACTN_COMPLETED_OK, ))

    def ftp_RNTO(self, toName):
        fromName = self._fromName
        del self._fromName
        self.state = self.AUTHED
        try:
            fromsegs = toSegments(self.workingDirectory, fromName)
            tosegs = toSegments(self.workingDirectory, toName)
        except InvalidPath:
            return defer.fail(FileNotFoundError(fromName))
        # TODO:
        return succeed((REQ_FILE_ACTN_COMPLETED_OK, ))


#------------------------------------------------------------------------------


class BitDustFTPFactory(FTPFactory):
    protocol = BitDustFTP


#------------------------------------------------------------------------------

if __name__ == '__main__':
    lg.set_debug_level(20)
    settings.init()
    backup_fs.init()
    backup_control.init()
    # print backup_fs.IsDir('/Users')
    init()
    reactor.run()  # @UndefinedVariable
    settings.shutdown()

#     def ftp_NLST(self, path):
#         """
#         """
#         print 'ftp_NLST:', path
#         # XXX: why is this check different from ftp_RETR/ftp_STOR? See #4180
#         if self.dtpInstance is None or not self.dtpInstance.isConnected:
#             return defer.fail(
#                 BadCmdSequenceError('must send PORT or PASV before RETR'))
#
#         try:
#             segments = toSegments(self.workingDirectory, path)
#         except InvalidPath:
#             return defer.fail(FileNotFoundError(path))
#
#         def cbList(results, glob):
#             """
#             Send, line by line, each matching file in the directory listing, and
#             then close the connection.
#
#             @type results: A C{list} of C{tuple}. The first element of each
#                 C{tuple} is a C{str} and the second element is a C{list}.
#             @param results: The names of the files in the directory.
#
#             @param glob: A shell-style glob through which to filter results (see
#                 U{http://docs.python.org/2/library/fnmatch.html}), or L{None}
#                 for no filtering.
#             @type glob: L{str} or L{None}
#
#             @return: A C{tuple} containing the status code for a successful
#                 transfer.
#             @rtype: C{tuple}
#             """
#             self.reply(DATA_CNX_ALREADY_OPEN_START_XFR)
#             for (name, ignored) in results:
#                 if not glob or (glob and fnmatch.fnmatch(name, glob)):
#                     name = self._encodeName(name)
#                     self.dtpInstance.sendLine(name)
#             self.dtpInstance.transport.loseConnection()
#             return (TXFR_COMPLETE_OK,)
#
#         def listErr(results):
#             """
#             RFC 959 specifies that an NLST request may only return directory
#             listings. Thus, send nothing and just close the connection.
#
#             @type results: L{Failure}
#             @param results: The L{Failure} wrapping a L{FileNotFoundError} that
#                 occurred while trying to list the contents of a nonexistent
#                 directory.
#
#             @returns: A C{tuple} containing the status code for a successful
#                 transfer.
#             @rtype: C{tuple}
#             """
#             self.dtpInstance.transport.loseConnection()
#             return (TXFR_COMPLETE_OK,)
#
#         if _isGlobbingExpression(segments):
#             # Remove globbing expression from path
#             # and keep to be used for filtering.
#             glob = segments.pop()
#         else:
#             glob = None
#
#         d = Deferred()
#         d.addCallback(cbList, glob)
#         d.addErrback(listErr)
# #         if not driver.is_on('service_restores'):
# #             d.errback(None)
#         result = [(itm['id'], []) for itm in api.backups_list()['result']]
#         d.callback(result)
#         return d
